from typing import Optional, List
import numpy as np
import pandas as pd
import igraph
from anndata import AnnData
from functools import partial
from statsmodels.stats.weightstats import DescrStatsW

from .utils import get_X
from .. import logging as logg


def slide_cells(
    adata: AnnData,
    root_milestone,
    milestones,
    win: int = 50,
    mapping: bool = True,
    copy: bool = False,
    ext: bool = False,
):

    """\
    Assign cells in a probabilistic manner to non-intersecting windows along pseudotime.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    win
        number of cell per local pseudotime window.
    mapping
        project cells onto tree pseudotime in a probabilistic manner.
    copy
        Return a copy instead of writing to adata.
    ext
        Output the list externally instead of writting to anndata

    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns subsetted or else subset (keeping only
        significant features) and add fields to `adata`:

        `.uns['root_milestone->milestoneA<>milestoneB']['cell_freq']`
            List of np.array containing probability assignment of cells on non intersecting windows.

    """

    adata = adata.copy() if copy else adata

    graph = adata.uns["graph"]

    uns_temp = adata.uns.copy()

    # weird fix to avoid loss of milestone colors...
    if "milestones_colors" in adata.uns:
        mlsc = adata.uns["milestones_colors"].copy()

    dct = graph["milestones"]
    keys = np.array(list(dct.keys()))
    vals = np.array(list(dct.values()))

    leaves = list(map(lambda leave: dct[leave], milestones))
    root = dct[root_milestone]

    def getsegs(g, root, leave, graph):
        path = np.array(g.vs[:]["name"])[
            np.array(g.get_shortest_paths(str(root), str(leave)))
        ][0]
        segs = list()
        for i in range(len(path) - 1):
            segs = segs + [
                np.argwhere(
                    (
                        graph["pp_seg"][["from", "to"]]
                        .astype(str)
                        .apply(lambda x: all(x.values == path[[i, i + 1]]), axis=1)
                    ).to_numpy()
                )[0][0]
            ]
        segs = graph["pp_seg"].index[segs].tolist()
        return segs

    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
    img = igraph.Graph()
    img.add_vertices(
        np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
    )
    img.add_edges(edges)

    paths = list(map(lambda l: getsegs(img, root, l, graph), leaves))
    pp_probs = adata.obsm["X_R"].sum(axis=0)

    if len(milestones) == 2:
        seg_progenies = list(set.intersection(*[set(path) for path in paths]))
        seg_branch1 = list(set.difference(set(paths[0]), set(seg_progenies)))
        seg_branch2 = list(set.difference(set(paths[1]), set(seg_progenies)))

        pps = (
            graph["pp_info"]
            .PP[
                graph["pp_info"].seg.isin(
                    np.array(seg_progenies + seg_branch1 + seg_branch2).astype(str)
                )
            ]
            .index
        )

        seg_branch1 = [str(seg) for seg in seg_branch1]
        seg_branch2 = [str(seg) for seg in seg_branch2]
        seg_progenies = [str(seg) for seg in seg_progenies]
        segs_cur = np.unique(
            np.array(seg_progenies + seg_branch1 + seg_branch2).flatten().astype(str)
        )

    elif len(milestones) == 1:
        pps = (
            graph["pp_info"]
            .PP[graph["pp_info"].seg.isin(np.array(paths[0]).astype(str))]
            .index
        )
        seg_progenies = [str(seg) for seg in paths[0]]
        segs_cur = np.array(seg_progenies)

    def region_extract(pt_cur, segs_cur, nbranch):
        freq = list()

        pp_next = pps[
            (graph["pp_info"].loc[pps, "time"].values >= pt_cur)
            & graph["pp_info"].loc[pps, "seg"].isin(segs_cur).values
        ]

        cmsm = np.cumsum(
            pp_probs[pp_next][np.argsort(graph["pp_info"].loc[pp_next, "time"].values)]
        )
        inds = np.argwhere(cmsm > win).flatten()

        if len(inds) == 0:
            if cmsm.max() > win / 2:
                if mapping:
                    cell_probs = adata.obsm["X_R"][:, pp_next].sum(axis=1)
                else:
                    cell_probs = (
                        np.isin(
                            np.apply_along_axis(
                                lambda x: np.argmax(x), axis=1, arr=adata.obsm["X_R"]
                            ),
                            pp_next,
                        )
                        * 1
                    )
                freq = freq + [cell_probs]
            return freq
        else:
            pps_region = pp_next[
                np.argsort(graph["pp_info"].loc[pp_next, "time"].values)
            ][: inds[0]]
            if mapping:
                cell_probs = adata.obsm["X_R"][:, pps_region].sum(axis=1)
            else:
                cell_probs = (
                    np.isin(
                        np.apply_along_axis(
                            lambda x: np.argmax(x), axis=1, arr=adata.obsm["X_R"]
                        ),
                        pps_region,
                    )
                    * 1
                )

            freq = freq + [cell_probs]
            pt_cur = graph["pp_info"].loc[pps_region, "time"].max()

            if nbranch == 1:
                if (
                    sum(~graph["pp_info"].loc[pps_region, :].seg.isin(seg_progenies))
                    == 0
                ):
                    res = region_extract(pt_cur, segs_cur, nbranch)
                    return freq + res

            if nbranch == 2:
                if (
                    sum(~graph["pp_info"].loc[pps_region, :].seg.isin(seg_progenies))
                    == 0
                ):
                    res = region_extract(pt_cur, segs_cur, nbranch)
                    return freq + res

                elif (
                    sum(~graph["pp_info"].loc[pps_region, :].seg.isin(seg_branch1)) == 0
                ):
                    res = region_extract(pt_cur, segs_cur, nbranch)
                    return freq + res

                elif (
                    sum(~graph["pp_info"].loc[pps_region, :].seg.isin(seg_branch2)) == 0
                ):

                    res = region_extract(pt_cur, segs_cur, nbranch)
                    return freq + res

                elif ~(
                    sum(
                        ~graph["pp_info"]
                        .loc[pps_region, :]
                        .seg.isin([str(seg) for seg in seg_progenies])
                    )
                    == 0
                ):
                    pt_cur1 = (
                        graph["pp_info"]
                        .loc[pps_region, "time"][
                            graph["pp_info"]
                            .loc[pps_region, "seg"]
                            .isin([str(seg) for seg in seg_branch1])
                        ]
                        .max()
                    )
                    segs_cur1 = seg_branch1
                    pt_cur2 = (
                        graph["pp_info"]
                        .loc[pps_region, "time"][
                            graph["pp_info"]
                            .loc[pps_region, "seg"]
                            .isin([str(seg) for seg in seg_branch2])
                        ]
                        .max()
                    )
                    segs_cur2 = seg_branch2
                    res1 = region_extract(pt_cur1, segs_cur1, nbranch)
                    res2 = region_extract(pt_cur2, segs_cur2, nbranch)
                    return freq + res1 + res2

    pt_cur = graph["pp_info"].loc[pps, "time"].min()

    freq = region_extract(pt_cur, segs_cur, len(milestones))
    name = root_milestone + "->" + "<>".join(milestones)

    adata.uns = uns_temp

    freqs = list(map(lambda f: pd.Series(f, index=adata.obs_names), freq))

    if ext is False:
        if name in adata.uns:
            adata.uns[name]["cell_freq"] = freqs
        else:
            adata.uns[name] = {"cell_freq": freqs}
        logg.hint(
            "added \n"
            "    .uns['"
            + name
            + "']['cell_freq'], probability assignment of cells on "
            + str(len(freq))
            + " non intersecting windows."
        )

    if copy:
        return adata
    elif ext:
        return freqs
    else:
        None


def slide_cors(
    adata: AnnData,
    root_milestone,
    milestones: List,
    genesetA=None,
    genesetB=None,
    perm=False,
    layer: Optional[str] = None,
    copy: bool = False,
):
    """\
    Obtain gene module correlations in the non-intersecting windows along pseudotime.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    layer
        adata layer from which to compute the correlations.
    copy
        Return a copy instead of writing to adata.

    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns subsetted or else subset (keeping only
        significant features) and add fields to `adata`:

        `.uns['root_milestone->milestoneA<>milestoneB']['corAB']`
            Dataframe containing gene-gene correlation modules.

    """

    adata = adata.copy() if copy else adata

    graph = adata.uns["graph"]

    uns_temp = adata.uns.copy()

    # weird fix to avoid loss of milestone colors...
    if "milestones_colors" in adata.uns:
        mlsc = adata.uns["milestones_colors"].copy()

    dct = graph["milestones"]
    keys = np.array(list(dct.keys()))
    vals = np.array(list(dct.values()))

    leaves = list(map(lambda leave: dct[leave], milestones))
    root = dct[root_milestone]

    name = root_milestone + "->" + "<>".join(milestones)

    if (genesetA is None or genesetB is None) and len(milestones) == 1:
        raise ValueError(
            "You need two list of genes when a non-bifurcating trajectory is analysed!"
        )

    if genesetA is None or genesetB is None:
        bif = adata.uns[name]["fork"]
    freqs = adata.uns[name]["cell_freq"]
    nwin = len(freqs)

    if genesetA is None:
        genesetA = bif.index[
            (bif["branch"] == milestones[0]).values & (bif["module"] == "early").values
        ]
    if genesetB is None:
        genesetB = bif.index[
            (bif["branch"] == milestones[1]).values & (bif["module"] == "early").values
        ]

    genes = adata.var_names
    if perm:
        genesetA = np.random.choice(genes[~genes.isin(genesetA)], len(genesetA))
        genesetB = np.random.choice(genes[~genes.isin(genesetB)], len(genesetB))

    genesets = np.concatenate([genesetA, genesetB])

    X = get_X(adata, adata.obs_names, genesets, layer)

    X = pd.DataFrame(X, index=adata.obs_names, columns=genesets)
    X_r = X.rank(axis=0)

    def gather_cor(i, geneset):
        freq = freqs[i][adata.obs_names]
        cormat = pd.DataFrame(
            DescrStatsW(X_r.values, weights=freq).corrcoef,
            index=genesets,
            columns=genesets,
        )
        np.fill_diagonal(cormat.values, np.nan)
        return cormat.loc[:, geneset].mean(axis=1)

    gather = partial(gather_cor, geneset=genesetA)
    corA = pd.concat(list(map(gather, range(nwin))), axis=1)
    corA.columns = [str(c) for c in corA.columns]
    corA = dict(
        zip(
            ["genesetA", "genesetB"],
            [corA.loc[geneset] for geneset in [genesetA, genesetB]],
        )
    )

    gather = partial(gather_cor, geneset=genesetB)
    corB = pd.concat(list(map(gather, range(nwin))), axis=1)
    corB.columns = [str(c) for c in corB.columns]
    corB = dict(
        zip(
            ["genesetA", "genesetB"],
            [corB.loc[geneset] for geneset in [genesetA, genesetB]],
        )
    )

    groups = ["A", "B"] if len(milestones) == 1 else milestones

    adata.uns = uns_temp
    adata.uns[name]["corAB"] = dict(zip(groups, [corA, corB]))

    logg.hint(
        "added \n" "    .uns['" + name + "']['corAB'], gene-gene correlation modules."
    )

    return adata if copy else None



from numba import cuda, njit, prange
import math
from tqdm import tqdm
from joblib import Parallel
from scipy import sparse
import numpy as np
import shutil
from .. import logging as logg


def get_X(adata, cells, genes, layer, togenelist=False):
    if layer is None:
        if sparse.issparse(adata.X):
            X = adata[cells, genes].X.A
        else:
            X = adata[cells, genes].X
    else:
        if sparse.issparse(adata.layers[layer]):
            X = adata[cells, genes].layers[layer].A
        else:
            X = adata[cells, genes].layers[layer]

    if togenelist:
        return X.T.tolist()
    else:
        return X


class ProgressParallel(Parallel):
    def __init__(
        self, use_tqdm=True, total=None, file=None, desc=None, *args, **kwargs
    ):
        self._use_tqdm = use_tqdm
        self._total = total
        self._desc = desc
        self._file = file
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        with tqdm(
            disable=not self._use_tqdm,
            total=self._total,
            desc=self._desc,
            file=self._file,
        ) as self._pbar:
            return Parallel.__call__(self, *args, **kwargs)

    def print_progress(self):
        if self._total is None:
            self._pbar.total = self.n_dispatched_tasks
        self._pbar.n = self.n_completed_tasks
        self._pbar.refresh()


@cuda.jit
def process_R_gpu(R, sigma):
    x, y = cuda.grid(2)
    if x < R.shape[0] and y < R.shape[1]:
        R[x, y] = math.exp(-R[x, y] / sigma)


@cuda.jit
def norm_R_gpu(R, Rsum):
    x, y = cuda.grid(2)
    if x < R.shape[0] and y < R.shape[1]:
        R[x, y] = R[x, y] / Rsum[x]
        if math.isnan(R[x, y]):
            R[x, y] = 0


@cuda.jit
def matmul(A, B, C):
    i, j = cuda.grid(2)
    if i < C.shape[0] and j < C.shape[1]:
        tmp = 0.0
        for k in range(A.shape[1]):
            tmp += A[i, k] * B[k, j]
        C[i, j] = tmp


@njit(parallel=True)
def process_R_cpu(R, sigma):
    Rshape = R.shape
    R = R.ravel()
    for x in prange(len(R)):
        R[x] = math.exp(-R[x] / sigma)
    R.reshape(Rshape)


@njit(parallel=True)
def norm_R_cpu(R, Rsum):
    for x in prange(len(Rsum)):
        for y in range(R.shape[1]):
            R[x, y] = R[x, y] / Rsum[x]
            if math.isnan(R[x, y]):
                R[x, y] = 0


def cor_mat_cpu(A, B):
    import numpy as np

    A1 = A - A.mean(axis=0)
    B1 = B - B.mean(axis=0)
    res = (B1.T.dot(A1)).T / np.sqrt(
        (A1 ** 2).sum(axis=0).reshape(A1.shape[1], 1)
        @ (B1 ** 2).sum(axis=0).reshape(1, B1.shape[1])
    )
    return res.T


def cor_mat_gpu(A, B):
    import cupy as cp

    A1 = A - A.mean(axis=0)
    B1 = B - B.mean(axis=0)
    res = (B1.T.dot(A1)).T / cp.sqrt(
        (A1 ** 2).sum(axis=0).reshape(A1.shape[1], 1)
        @ (B1 ** 2).sum(axis=0).reshape(1, B1.shape[1])
    )
    return res.T


def mst_gpu(d):
    import numpy as np
    import cugraph
    import cudf
    import cupy as cp
    from cupyx.scipy.sparse.csr import csr_matrix as csr_cupy
    from cupyx.scipy.sparse.coo import coo_matrix
    from cugraph.tree.minimum_spanning_tree_wrapper import mst_double, mst_float
    import scipy

    csr_gpu = csr_cupy(d)
    offsets = cudf.Series(csr_gpu.indptr)
    indices = cudf.Series(csr_gpu.indices)

    num_verts = csr_gpu.shape[0]
    num_edges = len(csr_gpu.indices)
    weights = cudf.Series(csr_gpu.data)

    if weights.dtype == np.float32:
        mst = mst_float(num_verts, num_edges, offsets, indices, weights)

    else:
        mst = mst_double(num_verts, num_edges, offsets, indices, weights)

    mst = csr_cupy(
        coo_matrix(
            (mst.weight.values, (mst.src.values, mst.dst.values)),
            shape=(num_verts, num_verts),
        )
    ).get()
    return csr_cupy(scipy.sparse.triu(mst))


def getpath(g, root, tips, tip, tree, df):
    import warnings
    import numpy as np

    wf = warnings.filters.copy()
    warnings.filterwarnings("ignore")
    try:
        path = np.array(g.vs[:]["name"])[
            np.array(g.get_shortest_paths(str(root), str(tip)))
        ][0]
        segs = list()
        for i in range(len(path) - 1):
            segs = segs + [
                np.argwhere(
                    (
                        tree["pp_seg"][["from", "to"]]
                        .astype(str)
                        .apply(lambda x: all(x.values == path[[i, i + 1]]), axis=1)
                    ).to_numpy()
                )[0][0]
            ]
        segs = tree["pp_seg"].index[segs]
        pth = df.loc[df.seg.astype(int).isin(segs), :].copy(deep=True)
        pth["branch"] = str(root) + "_" + str(tip)
        warnings.filters = wf
        return pth
    except IndexError:
        pass


def palantir_on_seg(adata, seg, ms_data):
    import palantir

    adata_sub = adata[
        adata.obs.seg == seg,
    ]
    pr = palantir.core.run_palantir(
        ms_data.loc[adata_sub.obs_names, :], adata_sub.obs.t.idxmin()
    )
    return pr.pseudotime


@njit(parallel=True)
def get_SE(MSE, x, se):
    N = len(x)
    xmean = x.mean()
    xxmean = np.sum((x - xmean) ** 2)
    for i in range(N):
        se[i] = math.sqrt(MSE) * math.sqrt(1 + 1 / N + (x[i] - xmean) ** 2 / xxmean)


def bh_adjust(x, log=False):
    x.sort_values(ascending=True)
    if log:
        q = x.sort_values(ascending=True) + np.log(len(x) / (np.arange(len(x)) + 1))
    else:
        q = x.sort_values(ascending=True) * len(x) / (np.arange(len(x)) + 1)
    return (q.reindex(index=q.index[::-1]).cummin())[x.index]


def importeR(task):
    try:
        from rpy2.robjects import pandas2ri, Formula
        from rpy2.robjects.packages import PackageNotInstalledError, importr
        import rpy2.rinterface

        pandas2ri.activate()
        Rpy2 = True
    except ModuleNotFoundError as e:
        Rpy2 = (
            "rpy2 installation is necessary for "
            + task
            + '. \
            \nPlease use "pip3 install rpy2" to install rpy2'
        )
        Formula = False

    if not shutil.which("R"):
        R = (
            "R installation is necessary for "
            + task
            + ". \
            \nPlease install R and try again"
        )
    else:
        R = True

    try:
        rstats = importr("stats")
    except Exception as e:
        rstats = (
            "R installation is necessary for "
            + task
            + ". \
            \nPlease install R and try again"
        )

    try:
        rmgcv = importr("mgcv")
    except Exception as e:
        rmgcv = (
            'R package "mgcv" is necessary for '
            + task
            + ". \
            \nPlease install gam from https://cran.r-project.org/web/packages/gam/ and try again"
        )

    return Rpy2, R, rstats, rmgcv, Formula



"""
The following code has been translated from R package pagoda2, if you use any of these functions please cite:
Nikolas Barkas, Viktor Petukhov, Peter Kharchenko and Evan
Biederstedt (2021). pagoda2: Single Cell Analysis and Differential
Expression. R package version 1.0.2.
"""

from anndata import AnnData

import pandas as pd
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, issparse
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt
from scipy.stats import t
import statsmodels.api as sm


from .. import logging as logg
from .. import settings
from ..tools.utils import get_SE, bh_adjust, importeR


def filter_cells(
    adata: AnnData, device="cpu", p_level=None, subset=True, plot=False, copy=False
):

    """\
    Filter cells using on gene/molecule relationship.

    Code has been translated from pagoda2 R function gene.vs.molecule.cell.filter.


    Parameters
    ----------
    adata
        Annotated data matrix.
    device
        Run gene and molecule counting on either `cpu` or on `gpu`.
    p_level
        Statistical confidence level for deviation from the main trend, used for cell filtering (default=min(1e-3,1/adata.shape[0]))
    subset
        if False, add a column `outlier` in adata.obs, otherwise subset the adata.
    plot
        Plot the molecule distribution and the gene/molecule dependency fit.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------

    adata : anndata.AnnData
        if `copy=True` and `subset=True` it returns subsetted (removing outliers) or else add fields to `adata`:

        `.obs['outlier']`
            whether a cell is an outlier.

    """

    adata = adata.copy() if copy else adata

    logg.info("Filtering cells", reset=True)
    X = adata.X.copy()

    logg.info("    obtaining gene and molecule counts")
    if device == "cpu":
        log1p_total_counts = np.log1p(np.array(X.sum(axis=1))).ravel()
        X.data = np.ones_like(X.data)
        log1p_n_genes_by_counts = np.log1p(np.array(X.sum(axis=1))).ravel()
    elif device == "gpu":
        import cupy as cp
        from cupyx.scipy.sparse import csr_matrix as csr_matrix_gpu

        X = csr_matrix_gpu(X)
        log1p_total_counts = cp.log1p(X.sum(axis=1)).get().ravel()
        X.data = cp.ones_like(X.data)
        log1p_n_genes_by_counts = cp.log1p(X.sum(axis=1)).get().ravel()

    df = pd.DataFrame(
        {
            "log1p_total_counts": log1p_total_counts,
            "log1p_n_genes_by_counts": log1p_n_genes_by_counts,
        },
        index=adata.obs_names,
    )

    logg.info("    fitting RLM")

    rlm_model = sm.RLM.from_formula(
        "log1p_n_genes_by_counts ~ log1p_total_counts",
        df,
    ).fit()

    p_level = min(1e-3, 1 / adata.shape[0]) if p_level is None else p_level

    SSE_line = ((df.log1p_n_genes_by_counts - rlm_model.predict()) ** 2).sum()
    MSE = SSE_line / df.shape[0]
    z = t.ppf((p_level / 2, 1 - p_level / 2), df.shape[0])

    se = np.zeros(df.shape[0])
    get_SE(MSE, df.log1p_total_counts.values, se)
    pr = pd.DataFrame(
        {
            0: rlm_model.predict(),
            1: rlm_model.predict() + se * z[0],
            2: rlm_model.predict() + se * z[1],
        },
        index=adata.obs_names,
    )

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")

    outlier = (df.log1p_n_genes_by_counts < pr[1]) | (
        df.log1p_n_genes_by_counts > pr[2]
    )

    if plot:
        fig, ax = plt.subplots()
        idx = df.sort_values("log1p_total_counts").index
        ax.fill_between(
            df.log1p_total_counts[[idx[0], idx[-1]]],
            pr[1][[idx[0], idx[-1]]],
            pr[2][[idx[0], idx[-1]]],
            color="yellow",
            alpha=0.3,
        )
        df.loc[~outlier].plot.scatter(
            x="log1p_total_counts", y="log1p_n_genes_by_counts", c="k", ax=ax, s=1
        )
        df.loc[outlier].plot.scatter(
            x="log1p_total_counts", y="log1p_n_genes_by_counts", c="grey", ax=ax, s=1
        )

    if subset:
        adata._inplace_subset_obs(adata.obs_names[~outlier])
        logg.hint("subsetted adata.")

    else:
        adata.obs["outlier"] = outlier
        logg.hint("added \n" "    .obs['outlier'], boolean column indicating outliers.")

    return adata if copy else None


def batch_correct(
    adata, batch_key, layer="X", depth_scale=1e3, device="cpu", inplace=True
):

    """\
    batch correction of the count matrix.

    Code has been translated from pagoda2 R function setCountMatrix (plain model).

    Parameters
    ----------
    adata
        Annotated data matrix.
    batch_key
        Column name to use for batch.
    layer
        Which layer to correct, if layer doesn't exist, then correct X and save to layer
    depth_scale
        Depth scale.
    device
        Run method on either `cpu` or on `gpu`.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    adata : anndata.AnnData
        if `inplace=False` it returns the corrected matrix, else it update field to `adata`:

        `.X`
            batch-corrected count matrix.

    """
    if adata.is_view:
        adata._init_as_actual(adata.copy())

    if layer in adata.layers:
        X = adata.layers[layer].copy()
    else:
        X = adata.X.copy()

    logg.info("Performing pagoda2 batch correction", reset=True)
    if adata.obs[batch_key].dtype.name != "category":
        adata.obs[batch_key] = adata.obs[batch_key].astype("category")
    batches = adata.obs[batch_key].cat.categories
    nbatches = len(batches)

    if device == "cpu":
        depth = X.sum(axis=1)
        depth = np.array(depth).ravel()

        gene_av = (np.array(X.sum(axis=0)).ravel() + len(batches)) / (
            depth.sum() + len(batches)
        )
        tc = np.vstack([X[adata.obs[batch_key] == b, :].sum(axis=0) for b in batches])
        tc = np.log(tc + 1) - np.log(
            np.array([depth[adata.obs[batch_key].values == b].sum() for b in batches])
            + 1
        ).reshape(-1, 1)
        bc = np.exp(tc - np.log(gene_av.astype(np.float64)))
        bc = pd.DataFrame(np.transpose(bc), columns=batches)
        X = csr_matrix(X.transpose())

        batch = adata.obs[batch_key].cat.rename_categories(range(nbatches))
        count_gene = np.repeat(np.arange(X.shape[0]), np.diff(X.indptr))
        acc = np.transpose(np.vstack([count_gene, batch[X.indices].values]))
        X.data = X.data / bc.values[acc[:, 0], acc[:, 1]]
        logg.info("    depth scaling")
        X = X.transpose()
        d = depth / depth_scale
        X = X.multiply(1.0 / d[None, :].T)

    elif device == "gpu":
        import cupy as cp
        import cudf
        from cupyx.scipy.sparse import csr_matrix as csr_matrix_gpu

        X = csr_matrix_gpu(X)
        depth = X.sum(axis=1)
        depth = cp.array(depth).ravel()

        gene_av = (cp.array(X.sum(axis=0)).ravel() + len(batches)) / (
            depth.sum() + len(batches)
        )
        tc = cp.vstack([X[adata.obs[batch_key] == b, :].sum(axis=0) for b in batches])
        tc = cp.log(tc + 1) - cp.log(
            cp.array([depth[adata.obs[batch_key].values == b].sum() for b in batches])
            + 1
        ).reshape(-1, 1)
        bc = cp.exp(tc - np.log(gene_av.astype(cp.float64)))
        bc = cudf.DataFrame(np.transpose(bc.get()), columns=batches)
        X = csr_matrix_gpu(X.transpose())

        batch = adata.obs[batch_key].cat.rename_categories(range(nbatches))
        count_gene = cp.repeat(cp.arange(X.shape[0]), cp.diff(X.indptr).get().tolist())
        batch_to_stack = cp.array(batch.values[X.indices.get()])
        acc = cp.transpose(cp.vstack([count_gene, batch_to_stack]))
        X.data = X.data / bc.values[acc[:, 0], acc[:, 1]]
        X = X.transpose()
        logg.info("    depth scaling")
        d = depth / depth_scale
        X = X.multiply(1.0 / d[None, :].T)
        X = X.get()

    X = csr_matrix(X)
    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")

    if inplace:
        if layer == "X":
            adata.X = X
            logg.hint("updated \n" "    .X, batch corrected matrix.")
        else:
            adata.layers[layer] = X
            logg.hint(
                "updated \n" "    .layer['" + layer + "'], batch corrected matrix."
            )

    else:
        return csr_matrix(X)


def find_overdispersed(
    adata,
    gam_k: int = 5,
    alpha: float = 5e-2,
    layer: str = "X",
    plot: bool = False,
    copy: bool = False,
):

    """\
    Find overdispersed gene, using pagoda2 strategy.

    Code has been translated from pagoda2 R function adjustVariance.

    Parameters
    ----------
    adata
        Annotated data matrix.
    gam_k
        The k used for the generalized additive model.
    alpha
        The criterion used to measure statistical significance.
    layer
        Which layer to use.
    plot
        Plot selected genes.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

         .var['res']
             residuals of GAM fit.
         .var['lp']
             p-value.
         .var['lpa']
             BH adjusted p-value.
         .var['qv']
             percentile of qui-squared distribution.
         .var['highly_variable']
             feature is over-dispersed.

    """

    Rpy2, R, rstats, rmgcv, Formula = importeR("finding overdispersed features")

    if any(np.array(adata.X.sum(axis=0)).ravel() == 0):
        raise Exception(
            "Cannot find overdispersed features if some are not expressed in any cell!"
        )

    if not issparse(adata.X):
        raise Exception("Overdispersion requires sparse matrix!")

    logg.info("Finding overdispersed features", reset=True)

    adata = adata.copy() if copy else adata

    if layer == "X":
        X = adata.X.copy()
    else:
        X = adata.layers[layer].copy()

    logg.info("    computing mean and variances")

    m = np.log(np.array(X.mean(axis=0)).ravel())
    v = np.log(StandardScaler(with_mean=False).fit(X).var_)
    X.data = np.ones_like(X.data)
    n_obs = np.array(X.sum(axis=0)).ravel()

    df = pd.DataFrame({"m": m, "v": v, "n_obs": n_obs}, index=adata.var_names)

    logg.info("    gam fitting")
    m = rmgcv.gam(Formula("v~s(m,k=" + str(gam_k) + ")"), data=df)
    df["res"] = rstats.residuals(m, type="response")
    n_obs = df.n_obs

    df["lp"] = rstats.pf(np.exp(df.res), n_obs, n_obs, lower_tail=False, log_p=True)
    df["lpa"] = bh_adjust(df["lp"], log=True)

    n_cells = adata.shape[0]
    df["qv"] = (
        rstats.qchisq(df["lp"], n_cells - 1, lower_tail=False, log_p=True) / n_cells
    )

    ods = df["lpa"] < np.log(alpha)
    df["highly_variable"] = ods.values

    adata.var[df.columns] = df

    logg.info(
        "    found " + str(sum(df["highly_variable"])) + " over-dispersed features",
        time=True,
        end=" " if settings.verbosity > 2 else "\n",
    )
    logg.hint(
        "added \n"
        "    .var['res'], residuals of GAM fit.\n"
        "    .var['lp'], p-value.\n"
        "    .var['lpa'], BH adjusted p-value.\n"
        "    .var['qv'], percentile of qui-squared distribution.\n"
        "    .var['highly_variable'], feature is over-dispersed.\n"
    )

    if plot:
        fig, ax = plt.subplots()
        df.loc[~df["highly_variable"]].plot.scatter(
            x="m", y="v", ax=ax, c="lightgrey", s=1
        )
        df.loc[df["highly_variable"]].plot.scatter(x="m", y="v", ax=ax, s=1, c="k")
        ax.set_xlabel("log10(magnitude)")
        ax.set_ylabel("log10(variance)")

    return adata if copy else None



from typing import Optional

import numpy as np
import pandas as pd
from anndata import AnnData
import scipy.sparse as sp

from .. import logging as logg
from .. import settings

import phenograph


def cluster(
    adata: AnnData,
    knn: int = 10,
    metric: str = "euclidean",
    device: str = "cpu",
    copy: bool = False,
    n_jobs: int = 1,
):

    """\
    Cluster feature trends.

    The models are fit using *mgcv* R package. Note that since adata can currently only keep the
    same dimensions for each of its layers, the dataset is subsetted to keep only significant
    feratures.


    Parameters
    ----------
    adata
        Annotated data matrix.
    knn
        Number of neighbors.
    metric
        distance metric to use for clustering.
    device
        run the analysis on 'cpu' with phenograph, or on 'gpu' with grapheno.
    leaves
        restrain the fit to a subset of the tree (in combination with root).
    copy
        Return a copy instead of writing to adata.
    Returns
    -------

    adata : anndata.AnnData
        if `copy=True` it returns subsetted or else subset (keeping only
        significant features) and add fields to `adata`:

        `.var['fit_clusters']`
            cluster assignments for features.

    """

    adata = data.copy() if copy else adata

    if device == "gpu":
        from . import grapheno_modified

        logg.info("    clustering using grapheno")
        communities, G, Q = grapheno_modified.cluster(
            adata.layers["fitted"].T, metric=metric, n_neighbors=knn
        )
        communities = communities.get()
        G = sp.csr_matrix(G.to_numpy_array())

    elif device == "cpu":
        logg.info("    clustering using phenograph")
        communities, G, Q = phenograph.cluster(
            adata.layers["fitted"].T, primary_metric=metric, k=knn, n_jobs=n_jobs
        )
        G = sp.csr_matrix(G)

    adata.varp["similarity"] = G
    adata.var["fit_clusters"] = communities

    adata.uns["fit_cluster"] = dict(
        parameters=dict(knn=knn, metric=metric, device=device, n_jobs=n_jobs),
        modularity=Q,
    )

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "added\n"
        "    .obsp['similarity'], pairwise similarity graph.\n"
        "    .var['fit_clusters'], cluster assignments for features.\n"
        "    .uns['fit_clusters'], parameters and modularity of the clustering."
    )

    return adata if copy else None



from typing import Optional, Iterable
from typing_extensions import Literal
from anndata import AnnData
import numpy as np
import pandas as pd
import igraph
from ..tools.utils import get_X, getpath


def fork_stats(
    adata: AnnData,
    root_milestone: str,
    milestones: Iterable,
    module: Optional[str] = None,
    branch: Optional[str] = None,
):
    """\
    Extract statistics from the fork analysis.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    module
        subset features to a specific module.
    branch
        subset features to a specific milestone.

    Returns
    -------
    a :class:`pandas.DataFrame` extracted from adata.uns.

    """

    key = f"{root_milestone}->{milestones[0]}<>{milestones[1]}"
    df = adata.uns[key]["fork"]

    if module is not None:
        df = df.loc[df.module == module]
    if branch is not None:
        df = df.loc[df.branch == branch]

    return df


def slide_cors(
    adata: AnnData,
    root_milestone: str,
    milestones: Iterable,
    branch: str,
    geneset_branch: str,
):
    """\
    Extract statistics from the sliding window correlation analysis.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    branch
        subset features to a specific milestone.
    geneset_branch
        which geneset to show correlations.

    Returns
    -------
    a :class:`pandas.DataFrame` extracted from adata.uns.

    """

    key = f"{root_milestone}->{milestones[0]}<>{milestones[1]}"

    dct = dict(zip(milestones, ["genesetA", "genesetB"]))
    geneset_branch = dct[geneset_branch]
    corAB = pd.DataFrame(adata.uns[key]["corAB"])

    return corAB.loc[geneset_branch].loc[branch]


def modules(
    adata: AnnData,
    root_milestone: str,
    milestones: Iterable,
    layer: Optional[str] = None,
    module: Literal["early", "late", "all"] = "all",
):
    """\
    Extract mean expression of identified early and late modules.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    layer
        layer to use to calculate the mean.
    module
        extract either the early, late or both modules.

    Returns
    -------
    a :class:`pandas.DataFrame`.

    """

    graph = adata.uns["graph"]

    dct = graph["milestones"]

    leaves = list(map(lambda leave: dct[leave], milestones))
    root = dct[root_milestone]

    name = root_milestone + "->" + "<>".join(milestones)

    stats = adata.uns[name]["fork"]

    if "milestones_colors" not in adata.uns or len(adata.uns["milestones_colors"]) == 1:
        from . import palette_tools

        palette_tools._set_default_colors_for_categorical_obs(adata, "milestones")

    df = adata.obs.copy(deep=True)
    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
    img = igraph.Graph()
    img.add_vertices(
        np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
    )
    img.add_edges(edges)

    cells = np.unique(
        np.concatenate(
            [
                getpath(
                    img, root, adata.uns["graph"]["tips"], leaves[0], graph, df
                ).index,
                getpath(
                    img, root, adata.uns["graph"]["tips"], leaves[1], graph, df
                ).index,
            ]
        )
    )
    X = pd.DataFrame(
        get_X(adata, cells, stats.index, layer), index=cells, columns=stats.index
    )

    early_1 = (stats.branch.values == milestones[0]) & (stats.module.values == "early")
    late_1 = (stats.branch.values == milestones[0]) & (stats.module.values == "late")

    early_2 = (stats.branch.values == milestones[1]) & (stats.module.values == "early")
    late_2 = (stats.branch.values == milestones[1]) & (stats.module.values == "late")

    X_early = pd.DataFrame(
        {
            "early_" + milestones[0]: X.loc[:, early_1].mean(axis=1),
            "early_" + milestones[1]: X.loc[:, early_2].mean(axis=1),
        },
        index=X.index,
    )

    X_late = pd.DataFrame(
        {
            "late_" + milestones[0]: X.loc[:, late_1].mean(axis=1),
            "late_" + milestones[1]: X.loc[:, late_2].mean(axis=1),
        },
        index=X.index,
    )

    if module == "all":
        return pd.concat([X_early, X_late], axis=1)
    elif module == "early":
        return X_early
    else:
        return X_late



import numpy as np
import matplotlib.pyplot as plt
from anndata import AnnData
from typing import Optional, Union
from scanpy.plotting._utils import savefig_or_show


def linearity_deviation(
    adata: AnnData,
    start_milestone,
    end_milestone,
    ntop_genes: int = 30,
    fontsize: int = 8,
    show: Optional[bool] = None,
    save: Union[str, bool, None] = None,
):

    """\
    Plot the results generated by `tl.linearity_deviation`.

    Parameters
    ----------
    adata
        Annotated data matrix.
    start_milestone
        tip defining the starting point of analysed segment.
    end_milestone
        tip defining the end point of analysed segment.
    ntop_genes
        number of top genes to show.
    fontsize
        Fontsize for gene names.
    show
        show the plot.
    save
        save the plot.

    Returns
    -------
    If `show==False` a matrix of :class:`~matplotlib.axes.Axes`

    """

    name = start_milestone + "->" + end_milestone

    topgenes = adata.var[name + "_rss"].sort_values(ascending=False)[:ntop_genes]
    ymin = np.min(topgenes)
    ymax = np.max(topgenes)
    ymax += 0.3 * (ymax - ymin)

    fig, ax = plt.subplots()
    ax.set_ylim(ymin, ymax)
    ax.set_xlim(-0.9, len(topgenes) - 0.1)
    for ig, gene_name in enumerate(topgenes.index):
        ax.text(
            ig,
            topgenes[gene_name],
            gene_name,
            rotation="vertical",
            verticalalignment="bottom",
            horizontalalignment="center",
            fontsize=fontsize,
        )

    ax.set_xlabel("ranking")
    ax.set_ylabel("deviance from linearity")

    savefig_or_show("linearity_transition", show=show, save=save)

    if show == False:
        return ax



from anndata import AnnData


def test_association(adata: AnnData, log_A: bool = False):
    """\
    Plot a set of fitted features over pseudotime.

    Parameters
    ----------
    adata
        Annotated data matrix.
    log_A
        change the xaxis scale to log.

    Returns
    -------
    just the plot.

    """

    stats = adata.var.copy(deep=True)
    # correct for zeros but setting them to the lowest value
    stats.loc[stats.fdr == 0, "fdr"] = stats.fdr[stats.fdr != 0].min()
    colors = ["red" if signi else "grey" for signi in stats.signi.values]
    stats.plot.scatter(x="A", y="fdr", color=colors, logx=log_A, logy=True)



try:
    from importlib_metadata import version
except:
    from importlib.metadata import version
__version__ = version(__name__)
del version

from anndata import AnnData

from . import pp
from . import tl
from . import pl
from . import datasets
from . import get
from . import settings
from .settings import set_figure_pubready



import os

os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"

from typing import Union, Optional, Iterable

import numpy as np
import pandas as pd
from functools import partial, reduce
from anndata import AnnData
import shutil
import sys
import copy
import igraph
import warnings

from joblib import delayed
from tqdm import tqdm

from .. import logging as logg
from .. import settings
from .utils import getpath, get_X, importeR, ProgressParallel


Rpy2, R, rstats, rmgcv, Formula = importeR("fitting associated features")
check = [type(imp) == str for imp in [Rpy2, R, rstats, rmgcv, Formula]]


def fit(
    adata: AnnData,
    root=None,
    leaves=None,
    layer: Optional[str] = None,
    n_map: int = 1,
    n_jobs: int = 1,
    gamma: float = 1.5,
    save_raw: bool = True,
    copy: bool = False,
):

    """\
    Model feature expression levels as a function of tree positions.

    The models are fit using *mgcv* R package. Note that since adata can currently only keep the
    same dimensions for each of its layers. While the dataset is subsetted to keep only significant
    feratures, the unsubsetted dataset is kept in adata.raw (save_raw parameter).


    Parameters
    ----------
    adata
        Annotated data matrix.
    root
        restrain the fit to a subset of the tree (in combination with leaves).
    leaves
        restrain the fit to a subset of the tree (in combination with root).
    layer
        adata layer to use for the fitting.
    n_map
        number of cell mappings from which to do the test.
    n_jobs
        number of cpu processes used to perform the test.
    gamma
        stringency of penalty.
    saveraw
        save the unsubsetted anndata to adata.raw
    copy
        Return a copy instead of writing to adata.
    Returns
    -------

    adata : anndata.AnnData
        if `copy=True` it returns subsetted or else subset (keeping only
        significant features) and add fields to `adata`:

        `.layers['fitted']`
            fitted features on the trajectory for all mappings.

    """

    if any(check):
        idx = np.argwhere(
            [type(imp) == str for imp in [Rpy2, R, rstats, rmgcv, Formula]]
        ).min()
        raise Exception(np.array([Rpy2, R, rstats, rmgcv, Formula])[idx])

    adata = adata.copy() if copy else adata

    if "signi" not in adata.var.columns:
        raise ValueError(
            "You need to run `tl.test_association` before fitting features."
        )

    genes = adata.var_names[adata.var.signi]

    graph = adata.uns["graph"]
    tips = graph["tips"]

    mlsc_temp = None
    if leaves is not None:
        # weird hack to keep milestones colors saved
        if "milestones_colors" in adata.uns:
            mlsc = adata.uns["milestones_colors"].copy()
            mlsc_temp = mlsc.copy()
        dct = graph["milestones"]
        keys = np.array(list(dct.keys()))
        vals = np.array(list(dct.values()))

        leaves = list(map(lambda leave: dct[leave], leaves))
        root = dct[root]

    if root is None:
        root = graph["root"]
        tips = tips[~np.isin(tips, root)]
    root2 = None
    if "root2" in graph:
        root2 = graph["root2"]
        tips = tips[~np.isin(tips, graph["root2"])]

    if leaves is not None:
        tips = leaves

    logg.info("fit features associated with the trajectory", reset=True, end="\n")

    stat_assoc = list()

    for m in tqdm(
        range(n_map), disable=n_map == 1, file=sys.stdout, desc="    multi mapping "
    ):
        if "t_old" in adata.obs.columns:
            df = adata.obs.copy()
        else:
            df = adata.uns["pseudotime_list"][str(m)]
        edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
        img = igraph.Graph()
        img.add_vertices(
            np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
        )
        img.add_edges(edges)

        subtree = pd.concat(
            list(map(lambda tip: getpath(img, root, tips, tip, graph, df), tips)),
            axis=0,
        )
        if root2 is not None:
            subtree = pd.concat(
                [
                    subtree,
                    pd.concat(
                        list(
                            map(
                                lambda tip: getpath(img, root2, tips, tip, graph, df),
                                tips,
                            )
                        ),
                        axis=0,
                    ),
                ]
            )
        subtree = subtree[["t", "branch"]]
        subtree["gamma"] = gamma

        Xgenes = get_X(adata, subtree.index, genes, layer, togenelist=True)

        data = list(zip([subtree] * len(Xgenes), Xgenes))

        stat = ProgressParallel(
            n_jobs=n_jobs,
            total=len(data),
            use_tqdm=n_map == 1,
            file=sys.stdout,
            desc="    single mapping ",
        )(delayed(gt_fun)(data[d]) for d in range(len(data)))

        stat_assoc = stat_assoc + [stat]

    for i in range(len(stat_assoc)):
        stat_assoc[i] = pd.concat(stat_assoc[i], axis=1)
        stat_assoc[i].columns = adata.var_names[adata.var.signi]

    names = np.arange(len(stat_assoc)).astype(str).tolist()
    dictionary = dict(zip(names, stat_assoc))

    if n_map == 1:
        fitted = dictionary["0"]
    else:
        dfs = list(dictionary.values())
        fitted = reduce(lambda x, y: x.add(y, fill_value=0), dfs) / n_map

    if save_raw:
        adata.raw = adata

    adata._inplace_subset_obs(np.unique(dictionary["0"].index))
    adata._inplace_subset_var(genes)

    adata.layers["fitted"] = fitted.loc[adata.obs_names, :]

    if mlsc_temp is not None:
        adata.uns["milestones_colors"] = mlsc_temp

    logg.info(
        "    finished (adata subsetted to keep only fitted features!)",
        time=True,
        end=" " if settings.verbosity > 2 else "\n",
    )

    if save_raw:
        logg.hint(
            "added\n"
            "    .layers['fitted'], fitted features on the trajectory for all mappings.\n"
            "    .raw, unfiltered data."
        )
    else:
        logg.hint(
            "added\n"
            "    .layers['fitted'], fitted features on the trajectory for all mappings."
        )

    return adata if copy else None


def gt_fun(data):
    sdf = data[0]
    sdf["exp"] = data[1]
    gamma = sdf["gamma"][0]

    global rmgcv
    global rstats

    def gamfit(b):
        m = rmgcv.gam(
            Formula("exp~s(t,bs='ts')"),
            data=sdf.loc[sdf["branch"] == b, :],
            gamma=gamma,
        )
        return pd.Series(
            rmgcv.predict_gam(m), index=sdf.loc[sdf["branch"] == b, :].index
        )

    mdl = list(map(gamfit, sdf.branch.unique()))

    return pd.concat(mdl, axis=1, sort=False).apply(np.nanmean, axis=1)



from typing import Union, Sequence, Optional
import numpy as np
import pandas as pd
import igraph
import matplotlib.pyplot as plt
from scFates.tools.utils import get_X
import scanpy as sc
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.ticker import MaxNLocator
import matplotlib.cm as cm
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scanpy.plotting._utils import savefig_or_show
from anndata import AnnData


def matrix(
    adata: AnnData,
    features: Sequence,
    nbins: int = 5,
    layer: str = "fitted",
    annot_var: bool = False,
    annot_top: bool = True,
    link_seg: bool = True,
    feature_style: str = "normal",
    feature_spacing: float = 1,
    figsize: Union[None, tuple] = None,
    show: Optional[bool] = None,
    save: Union[str, bool, None] = None,
    **kwargs
):
    """\
    Plot a set of features as per-segment matrix plots of binned pseudotimes.

    Parameters
    ----------
    adata
        Annotated data matrix.
    features
        Name of the features.
    nbins
        Number of pseudotime bins per segment.
    layer
        Layer to use for the expression to display.
    annot_var
        Annotate overall tree amplitude of expression of the marker (from .var['A']).
    annot_top
        Display milestones gradient for each segment on top of plots.
    link_seg
        Link the segment together to keep track of the the tree progression.
    feature_style
        Font style of the feature labels.
    feature_spacing
        When figsize is None, controls the the height of each rows.
    figsize
        Custom figure size.
    show
        show the plot.
    save
        save the plot.
    save_genes
        save list of genes following the order displayed on the heatmap.
    **kwargs
        arguments passed to :class:`scanpy.pl.MatrixPlot`

    Returns
    -------
    If `show==False` an array of :class:`~matplotlib.axes.Axes`

    """
    adata = adata[:, features].copy()

    if (layer == "fitted") & ("fitted" not in adata.layers):
        print("Features not fitted, using X expression matrix instead")
        layer = None
    X = get_X(adata, adata.obs_names, adata.var_names, layer=layer)

    adata.X = X / X.max(axis=0).ravel()

    graph = adata.uns["graph"]

    dct = graph["milestones"]
    keys = np.array(list(dct.keys()))
    vals = np.array(list(dct.values()))

    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
    img = igraph.Graph(directed=True)
    img.add_vertices(vals.astype(str))
    img.add_edges(edges)

    allpaths = img.get_all_shortest_paths(
        str(graph["root"]), to=graph["tips"].astype(str)
    )

    allpaths = np.array(allpaths, dtype=object)[
        np.argsort(np.array([len(p) for p in allpaths]))
    ]

    order = allpaths[0]
    for i in range(1, len(allpaths)):
        order = order + np.array(allpaths[i])[~np.isin(allpaths[i], order)].tolist()

    order = np.array(order)[1:]

    order = pd.Series(graph["milestones"].keys(), index=graph["milestones"].values())[
        np.array(img.vs["name"])[order].astype(int)
    ]
    order = pd.Series(
        range(len(adata.obs.seg.cat.categories)), index=graph["pp_seg"]["to"]
    )[order.index].values

    vs2mils = pd.Series(dct.keys(), index=dct.values())

    cellsel = [
        adata.obs.milestones[adata.obs.seg == s]
        for s in adata.obs.seg.cat.categories[order]
    ]

    fig, axs = plt.subplots(
        1,
        len(order) + 1 * annot_var,
        constrained_layout=False,
        sharey=True,
        figsize=(
            len(order) * 0.85 + 0.85 + annot_var * 1,
            (len(features) + 0.8 * annot_top + 1 + 0.8 * annot_var)
            / (5 - feature_spacing),
        )
        if figsize is None
        else figsize,
    )

    pos = np.arange(len(order), 0, -1)
    caxs = []
    for i, s in enumerate(order):
        adata_sub = adata[adata.obs.seg == adata.obs.seg.cat.categories[s]].copy()
        adata_sub.obs["split"] = pd.cut(adata_sub.obs.t, bins=nbins)

        ss = int(adata.obs.seg.cat.categories[s])
        sel = (
            adata.obs.milestones.cat.categories
            == vs2mils[graph["pp_seg"].loc[ss]["from"]]
        )
        start = np.array(adata.uns["milestones_colors"])[sel][0]
        sel = (
            adata.obs.milestones.cat.categories
            == vs2mils[graph["pp_seg"].loc[ss]["to"]]
        )
        end = np.array(adata.uns["milestones_colors"])[sel][0]

        my_cm = LinearSegmentedColormap.from_list("aspect", [start, end])

        if "use_raw" not in kwargs:
            kwargs["use_raw"] = False

        M = sc.pl.MatrixPlot(adata_sub, features, "split", vmin=0, vmax=1, **kwargs)
        M.swap_axes()
        M._mainplot(axs[i])
        axs[i].set_xticklabels("")
        axs[i].set_xticks([])
        plt.margins(y=10)
        plt.setp(axs[i].get_yticklabels(), style=feature_style)
        if annot_top:
            divider = make_axes_locatable(axs[i])
            cax = divider.new_vertical(size=0.2, pad=0.05, pack_start=False)
            caxs.append(cax)
            mappable = cm.ScalarMappable(cmap=my_cm)

            fig.add_axes(cax)
            cbar = fig.colorbar(mappable, cax=cax, orientation="horizontal")
            cbar.set_ticks([])
            cbar.outline.set_linewidth(1.5)

    if annot_var:
        Amps = adata.var.loc[features, "A"]
        axs[i + 1].barh(
            np.arange(len(features)) + 0.5,
            Amps,
            color="salmon",
            height=0.65,
            left=0,
            edgecolor="black",
            label="A",
        )
        axs[i + 1].invert_yaxis()
        axs[i + 1].spines["top"].set_visible(False)
        axs[i + 1].spines["left"].set_visible(False)
        axs[i + 1].spines["right"].set_visible(False)

        axs[i + 1].get_xaxis().tick_bottom()
        axs[i + 1].tick_params(left=False)
        axs[i + 1].xaxis.set_major_locator(MaxNLocator(integer=True))
        axs[i + 1].grid(False)

        if annot_top:
            divider = make_axes_locatable(axs[i + 1])
            cax = divider.new_vertical(size=0.2, pad=0.05, pack_start=False)
            fig.add_axes(cax)
            cax.annotate("Amplitude", (0, 0), va="bottom", ha="left", size=12)
            cax.axis("off")

    if link_seg:
        caxs_dct = dict(zip([adata.obs.seg.cat.categories[o] for o in order], caxs))

        kw = dict(
            arrowprops=dict(
                arrowstyle="<|-",
                facecolor="k",
                connectionstyle="bar,fraction=.2",
                shrinkA=0.1,
            ),
            zorder=0,
            va="center",
            xycoords="axes fraction",
            annotation_clip=False,
        )

        kwclose = dict(
            arrowprops=dict(
                arrowstyle="<|-",
                facecolor="k",
                connectionstyle="bar,fraction=1",
                shrinkA=0.1,
            ),
            zorder=0,
            va="center",
            xycoords="axes fraction",
            annotation_clip=False,
        )

        pp_seg = adata.uns["graph"]["pp_seg"]

        dsts = pd.Series(range(len(caxs_dct)), index=caxs_dct.keys())
        for s, cax in caxs_dct.items():
            fro = pp_seg.loc[int(s), "from"]
            to = pp_seg.loc[int(s), "to"]
            for n in pp_seg.index[pp_seg["from"] == to]:
                if dsts[str(n)] - dsts[s] > 1:
                    cax.annotate("", xy=[1, 1], xytext=[2.4, 1], **kw)
                else:
                    cax.annotate("", xy=[1, 1], xytext=[1.2, 1], **kwclose)

    savefig_or_show("matrix", show=show, save=save)

    if show == False:
        return axs



from typing import Union, Optional, List, Iterable
from typing_extensions import Literal

import numpy as np
import pandas as pd
import igraph
from anndata import AnnData
from scipy import sparse
import itertools

from joblib import delayed
from functools import partial
from statsmodels.stats.weightstats import DescrStatsW
from skmisc.loess import loess
from scipy.stats import norm
from statsmodels.stats.multitest import multipletests

import warnings
from .. import logging as logg
from .. import settings
from .utils import getpath, ProgressParallel, get_X

import sys

sys.setrecursionlimit(10000)

from scFates.tools.utils import importeR

Rpy2, R, rstats, rmgcv, Formula = importeR("Fitting syncho path on cells")


def synchro_path(
    adata: AnnData,
    root_milestone,
    milestones,
    genesetA: Optional[Iterable] = None,
    genesetB: Optional[Iterable] = None,
    n_map=1,
    n_jobs=None,
    layer: Optional[str] = None,
    perm=True,
    w=200,
    step=30,
    winp=10,
    knots=10,
    copy: bool = False,
):
    """\
    Estimates pseudotime trends of local intra- and inter-module correlations of fates-specific modules.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    n_map
        number of probabilistic cells projection to use for estimates.
    n_jobs
        number of cpu processes to perform estimates (per mapping).
    layer
        adata layer to use for estimates.
    perm
        estimate control trends for local permutations instead of real expression matrix.
    w
        local window, in number of cells, to estimate correlations.
    step
        steps, in number of cells, between local windows.
    winp
        window of permutation in cells.
    knots
        number of knots for GAM fit of corAB on cells pre-fork
    copy
        Return a copy instead of writing to adata.

    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns subsetted or else subset (keeping only
        significant features) and add fields to `adata`:

        `.uns['root_milestone->milestoneA<>milestoneB']['synchro']`
            Dataframe containing mean local gene-gene correlations of all possible gene pairs inside one module, or between the two modules.
        `.obs['intercor root_milestone->milestoneA<>milestoneB']`
            loess fit of inter-module mean local gene-gene correlations prior to bifurcation

    """

    adata = adata.copy() if copy else adata

    logg.info("computing local correlations", reset=True)

    graph = adata.uns["graph"]

    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
    img = igraph.Graph()
    img.add_vertices(
        np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
    )
    img.add_edges(edges)

    uns_temp = adata.uns.copy()

    if "milestones_colors" in adata.uns:
        mlsc = adata.uns["milestones_colors"].copy()

    dct = graph["milestones"]
    keys = np.array(list(dct.keys()))
    vals = np.array(list(dct.values()))

    leaves = list(map(lambda leave: dct[leave], milestones))
    root = dct[root_milestone]

    name = root_milestone + "->" + "<>".join(milestones)

    if genesetA is None:
        bif = adata.uns[name]["fork"]
        genesetA = bif.index[(bif.module == "early") & (bif.branch == milestones[0])]
        genesetB = bif.index[(bif.module == "early") & (bif.branch == milestones[1])]

    genesets = np.concatenate([genesetA, genesetB])

    if n_map == 1:
        logg.info("    single mapping")

    def synchro_map(m):
        df = adata.uns["pseudotime_list"][str(m)]
        edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
        img = igraph.Graph()
        img.add_vertices(
            np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
        )
        img.add_edges(edges)

        def synchro_milestone(leave):
            cells = getpath(img, root, graph["tips"], leave, graph, df)
            cells = cells.sort_values("t").index

            X = get_X(adata, cells, genesets, layer)
            mat = pd.DataFrame(X, index=cells, columns=genesets)

            if permut == True:
                winperm = np.min([winp, mat.shape[0]])
                for i in np.arange(0, mat.shape[0] - winperm, winperm):
                    mat.iloc[i : (i + winperm), :] = (
                        mat.iloc[i : (i + winperm), :]
                        .apply(np.random.permutation, axis=0)
                        .values
                    )

            def slide_path(i):
                cls = mat.index[i : (i + w)]
                cor = mat.loc[cls, :].corr(method="spearman")
                corA = cor.loc[:, genesetA].mean(axis=1)
                corB = cor.loc[:, genesetB].mean(axis=1)
                corA[genesetA] = (
                    (corA[genesetA] - 1 / len(genesetA))
                    * len(genesetA)
                    / (len(genesetA) - 1)
                )
                corB[genesetB] = (
                    (corB[genesetB] - 1 / len(genesetB))
                    * len(genesetB)
                    / (len(genesetB) - 1)
                )

                return pd.Series(
                    {
                        "t": adata.obs.t[cls].mean(),
                        "dist": (corA[genesetA].mean() - corA[genesetB].mean()) ** 2
                        + (corB[genesetA].mean() - corB[genesetB].mean()) ** 2,
                        "corAA": corA[genesetA].mean(),
                        "corBB": corB[genesetB].mean(),
                        "corAB": corA[genesetB].mean(),
                        "n_map": m,
                    }
                )

            ww = np.arange(0, mat.shape[0] - w, step)

            res = ProgressParallel(
                n_jobs=n_jobs,
                total=len(ww),
                use_tqdm=n_map == 1,
                file=sys.stdout,
                desc="    to " + str(keys[vals == leave][0]),
            )(delayed(slide_path)(i) for i in ww)

            return pd.concat(res, axis=1).T

        return pd.concat(list(map(synchro_milestone, leaves)), keys=milestones)

    if n_map > 1:
        permut = False
        stats = ProgressParallel(
            n_jobs=n_jobs, total=n_map, file=sys.stdout, desc="    multi mapping"
        )(delayed(synchro_map)(i) for i in range(n_map))
        allcor_r = pd.concat(stats)
        if perm:
            permut = True

            stats = ProgressParallel(
                n_jobs=n_jobs,
                total=n_map,
                file=sys.stdout,
                desc="    multi mapping permutations",
            )(delayed(synchro_map)(i) for i in range(n_map))
            allcor_p = pd.concat(stats)
            allcor = pd.concat([allcor_r, allcor_p], keys=["real", "permuted"])
        else:
            allcor = pd.concat([allcor_r], keys=["real"])
    else:
        permut = False
        allcor_r = pd.concat(list(map(synchro_map, range(n_map))))

        if perm:
            permut = True
            allcor_p = pd.concat(list(map(synchro_map, range(n_map))))
            allcor = pd.concat([allcor_r, allcor_p], keys=["real", "permuted"])
        else:
            allcor = pd.concat([allcor_r], keys=["real"])

    runs = pd.DataFrame(allcor.to_records())["level_0"].unique()

    dct_cormil = dict(
        zip(
            ["corAA", "corBB", "corAB"],
            [m + "\nintra-module" for m in milestones]
            + [" vs ".join(milestones)[0] + "\ninter-module"],
        )
    )

    fork = list(
        set(img.get_shortest_paths(str(root), str(leaves[0]))[0]).intersection(
            img.get_shortest_paths(str(root), str(leaves[1]))[0]
        )
    )
    fork = np.array(img.vs["name"], dtype=int)[fork]
    fork_t = adata.uns["graph"]["pp_info"].loc[fork, "time"].max()
    res = allcor.loc[allcor.t < fork_t, :]
    res = res[~res.t.duplicated()]

    m = rmgcv.gam(
        Formula("corAB ~ s(t, bs = 'cs',k=%s)" % knots),
        data=res,
    )
    pred = rmgcv.predict_gam(m)

    tval = adata.obs.t.copy()
    tval[tval > fork_t] = np.nan

    def inter_values(tv):
        if ~np.isnan(tv):
            return pred[np.argmin(np.abs(res.t.values - tv))]
        else:
            return tv

    adata.obs["inter_cor " + name] = list(map(inter_values, tval))

    df = adata.uns["pseudotime_list"][str(0)]
    cells = np.concatenate(
        [
            getpath(img, root, graph["tips"], leaves[0], graph, df).index,
            getpath(img, root, graph["tips"], leaves[1], graph, df).index,
        ]
    )

    adata.obs.loc[~adata.obs_names.isin(cells), "inter_cor " + name] = np.nan

    adata.uns = uns_temp

    allcor = dict(
        zip(
            allcor.index.levels[0],
            [
                dict(
                    zip(
                        allcor.loc[l1].index.levels[0],
                        [
                            allcor.loc[l1].loc[l2]
                            for l2 in allcor.loc[l1].index.levels[0]
                        ],
                    )
                )
                for l1 in allcor.index.levels[0]
            ],
        )
    )

    if name in adata.uns:
        adata.uns[name]["synchro"] = allcor
    else:
        adata.uns[name] = {"synchro": allcor}

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "added \n"
        "    .uns['"
        + name
        + "']['synchro'], mean local gene-gene correlations of all possible gene pairs inside one module, or between the two modules.\n"
        "    .obs['inter_cor "
        + name
        + "'], GAM fit of inter-module mean local gene-gene correlations prior to bifurcation."
    )

    return adata if copy else None


def synchro_path_multi(
    adata: AnnData, root_milestone, milestones, copy=False, **kwargs
):
    """\
    Wrappers that call `tl.synchro_path` on the pairwise combination of all selected branches.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    kwargs
        arguments to pass to tl.synchro_path.

    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns subsetted or else subset (keeping only
        significant features) and add fields to `adata`:

        `.uns['root_milestone->milestoneA<>milestoneB']['synchro']`
            Dataframe containing mean local gene-gene correlations of all possible gene pairs inside one module, or between the two modules.
        `.obs['intercor root_milestone->milestoneA<>milestoneB']`
            loess fit of inter-module mean local gene-gene correlations prior to bifurcation

    """

    adata = adata.copy() if copy else adata

    logg.info("computing local correlations", reset=True)

    graph = adata.uns["graph"]

    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
    img = igraph.Graph()
    img.add_vertices(
        np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
    )
    img.add_edges(edges)

    uns_temp = adata.uns.copy()

    if "milestones_colors" in adata.uns:
        mlsc = adata.uns["milestones_colors"].copy()

    dct = graph["milestones"]
    keys = np.array(list(dct.keys()))
    vals = np.array(list(dct.values()))

    leaves = list(map(lambda leave: dct[leave], milestones))
    root = dct[root_milestone]

    name = root_milestone + "->" + "<>".join(milestones)

    bif = adata.uns[name]["fork"]

    genesets = dict(
        zip(
            milestones,
            [
                bif.index[(bif.module == "early") & (bif.branch == m)]
                for m in milestones
            ],
        )
    )

    pairs = list(itertools.combinations(milestones, 2))

    for m_pair in pairs:
        synchro_path(
            adata,
            root_milestone,
            m_pair,
            genesetA=genesets[m_pair[0]],
            genesetB=genesets[m_pair[1]],
            **kwargs
        )


def module_inclusion(
    adata,
    root_milestone,
    milestones,
    w: int = 300,
    step: int = 30,
    pseudotime_offset: Union["all", float] = "all",
    module: Literal["all", "early"] = "all",
    n_perm: int = 10,
    n_map: int = 1,
    map_cutoff: float = 0.8,
    n_jobs: int = 1,
    alp: int = 10,
    autocor_cut: float = 0.95,
    iterations: int = 15,
    parallel_mode: Literal["window", "mappings"] = "window",
    identify_early_features: bool = True,
    layer=None,
    perm: bool = False,
    winp: int = 10,
    copy: bool = False,
):
    """\
    Estimates the pseudotime onset of a feature within its fate-specific module.

    Parameters
    ----------
    adata
        Annotated data matrix.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    w
        local window, in number of cells, to estimate correlations.
    step
        steps, in number of cells, between local windows.
    pseudotime_offset
        restrict the cell selection up to a pseudotime offset after the fork
    module
        restrict the gene selection to already classified early genes.
    n_perm
        number of permutations used to estimate the background local correlations.
    n_map
        number of probabilistic cells projection to use for estimates.
    map_cutoff
        proportion of mapping in which inclusion pseudotimne was found for a given feature to keep it.
    n_jobs
        number of cpu processes to perform estimates.
    alp
        parameter regulating stringency of inclusion event.
    autocor_cut
        cutoff on correlation of inclusion times between sequential iterations of the algorithm to stop it.
    iterations
        maximum number of iterations of the algorithm.
    parallel_mode
        whether to run in parallel over the windows of cells or the mappings.
    identify_early_features
        classify a feature as early if its inclusion pseudotime is before the bifurcation
    layer
        adata layer to use for estimates.
    perm
        do local estimates for locally permuted expression matrix.
    winp
        window of permutation in cells.
    copy
        Return a copy instead of writing to adata.

    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns subsetted or else subset (keeping only
        significant features) and add fields to `adata`:

        `.uns['root_milestone->milestoneA<>milestoneB']['module_inclusion']`
            Dataframes ontaining inclusion timing for each gene (rows) in each probabilistic cells projection (columns).
        `.uns['root_milestone->milestoneA<>milestoneB']['fork']`
            Updated with 'inclusion' pseudotime column and 'module column if `identify_early_features=True`'
    """

    adata = adata.copy() if copy else adata

    logg.info("Calculating onset of features within their own module", reset=True)

    graph = adata.uns["graph"]

    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
    img = igraph.Graph()
    img.add_vertices(
        np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
    )
    img.add_edges(edges)

    uns_temp = adata.uns.copy()

    if "milestones_colors" in adata.uns:
        mlsc = adata.uns["milestones_colors"].copy()

    dct = graph["milestones"]
    keys = np.array(list(dct.keys()))
    vals = np.array(list(dct.values()))

    leaves = list(map(lambda leave: dct[leave], milestones))
    root = dct[root_milestone]

    name = root_milestone + "->" + "<>".join(milestones)

    common_seg = list(
        set.intersection(
            *list(
                map(
                    lambda l: set(img.get_shortest_paths(str(root), str(l))[0]),
                    leaves,
                )
            )
        )
    )
    common_seg = np.array(img.vs["name"], dtype=int)[common_seg]
    fork_t = adata.uns["graph"]["pp_info"].loc[common_seg, "time"].max()

    def onset_map(m):
        df = adata.uns["pseudotime_list"][str(m)]
        edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
        img = igraph.Graph()
        img.add_vertices(
            np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
        )
        img.add_edges(edges)

        def onset_milestone(milestone):
            sel = adata.uns[name]["fork"].branch == milestone
            sel = (
                sel & (adata.uns[name]["fork"].module == "early")
                if module == "early"
                else sel
            )
            geneset = adata.uns[name]["fork"].index[
                adata.uns[name]["fork"].branch == milestone
            ]
            cells = getpath(
                img,
                root,
                graph["tips"],
                adata.uns["graph"]["milestones"][milestone],
                graph,
                df,
            )

            if pseudotime_offset != "all":
                cells = cells.loc[cells.t < (fork_t + pseudotime_offset)]

            cells = cells.sort_values("t").index

            X = get_X(adata, cells, geneset, layer)
            mat = pd.DataFrame(X, index=cells, columns=geneset)

            if perm:
                winperm = np.min([winp, mat.shape[0]])
                for i in np.arange(0, mat.shape[0] - winperm, winperm):
                    mat.iloc[i : (i + winperm), :] = (
                        mat.iloc[i : (i + winperm), :]
                        .apply(np.random.permutation, axis=0)
                        .values
                    )

            mat = mat.T
            ww = np.arange(0, len(cells) - w, step)
            logW = pd.DataFrame(1, index=geneset, columns=range(len(ww)))

            def slide_cor(i):
                cls0 = mat.columns[ww[i] : min(ww[i] + w - 1, len(cells))]
                mat_1 = mat[cls0]
                mat_2 = mat_1.copy()
                mat_2[mat_2 != 0] = 1
                mat_2 = mat_2.dot(mat_2.T)
                mat_2[mat_2 < 10] = 0
                mat_2[mat_2 >= 10] = 1
                cor = mat_1.T.corr(method="spearman") * mat_2

                def perm_mat():
                    mat_perm = mat[cls0].apply(np.random.permutation, axis=0).values
                    mat_perm = pd.DataFrame(mat_perm, columns=cls0)
                    mat_perm_2 = mat_perm.copy()
                    mat_perm_2[mat_perm_2 != 0] = 1
                    mat_perm_2 = mat_perm_2.dot(mat_perm_2.T)
                    mat_perm_2[mat_perm_2 < 10] = 0
                    mat_perm_2[mat_perm_2 >= 10] = 1
                    return mat_perm.T.corr(method="spearman") * mat_perm_2

                allperm = [perm_mat() for i in range(n_perm)]

                return cor, allperm

            res = ProgressParallel(
                n_jobs=n_jobs,
                total=len(ww),
                use_tqdm=n_map == 1,
                file=sys.stdout,
                desc="    to " + milestone,
            )(delayed(slide_cor)(i) for i in range(len(ww)))

            cors = [r[0] for r in res]
            cors_p = [r[1] for r in res]

            def corEvol(cors, cors_p, logW):
                cor_Ps = []
                for i in range(len(cors)):
                    cor, cor_p = cors[i], cors_p[i]
                    corTrue = (cor.T * logW.iloc[:, i]).mean(axis=0)
                    cor_control = [
                        (cor_p.T.values * logW.iloc[:, i].values).mean(axis=0)
                        for cor_p in cor_p
                    ]
                    cor_Ps = cor_Ps + [
                        np.array(
                            [
                                sum(
                                    cor_c + np.random.uniform(0, 0.01, len(cor_c))
                                    >= corTrue
                                    for cor_c in cor_control
                                )
                            ]
                        )
                        / len(cor_control)
                    ]
                return cor_Ps

            logList = []
            autocor = 0

            def switch_point(r, alp):
                def logL(j):
                    v = 0
                    if j >= 2:
                        v = v
                    if j <= len(r):
                        v = v + sum(
                            -alp * r[j : len(r)] + np.log(alp / (1 - np.exp(-alp)))
                        )

                    return v

                switch_point = np.argmax([logL(j) for j in range(len(r) + 1)])
                n = len([logL(j) for j in range(len(r) + 1)])
                return np.concatenate(
                    [np.repeat(0, switch_point), np.repeat(1, n - switch_point - 1)]
                )

            i = 0
            auto_cor = 0
            while (autocor < autocor_cut) & (i <= iterations):
                pMat = np.vstack(corEvol(cors, cors_p, logW)).T
                sws = [switch_point(pMat[i, :], alp) for i in range(pMat.shape[0])]
                sws = np.vstack(sws)
                logW = pd.DataFrame(sws, index=geneset)
                logList = logList + [logW]

                if i > 0:
                    autocor = np.corrcoef(
                        logList[i - 1].idxmax(axis=1), logList[i].idxmax(axis=1)
                    )[1][0]

                i = i + 1

            incl_t = pd.Series(np.nan, index=logW.index)
            logW = logW.loc[logW.sum(axis=1) != 0]

            incl_t[logW.index] = [
                adata.obs.t[cells[(pos * (step - 1)) : (pos * (step - 1) + w)]].mean()
                for pos in logW.idxmax(axis=1)
            ]
            return incl_t

        return [onset_milestone(milestone) for milestone in milestones]

    n_jobs_map = 1
    if parallel_mode == "mappings":
        n_jobs_map = n_jobs
        n_jobs = 1

    stats = ProgressParallel(
        n_jobs=n_jobs_map,
        total=n_map,
        use_tqdm=n_map > 1,
        file=sys.stdout,
        desc="    multi mapping",
    )(delayed(onset_map)(i) for i in range(n_map))

    matSwitch = dict(
        zip(
            milestones,
            [
                pd.concat(
                    [s[i] for s in stats],
                    axis=1,
                    keys=np.arange(len([s[i] for s in stats])).astype(str),
                )
                for i in range(len(leaves))
            ],
        )
    )

    perm_str = "_perm" if perm else ""
    if perm:
        identify_early_features = False

    for m in milestones:
        props = 1 - np.isnan(matSwitch[m]).sum(axis=1) / n_map
        matSwitch[m] = matSwitch[m].loc[props > map_cutoff]

    adata.uns[name]["module_inclusion" + perm_str] = matSwitch

    updated = "."
    if perm == False:
        common_seg = list(
            set.intersection(
                *list(
                    map(
                        lambda l: set(img.get_shortest_paths(str(root), str(l))[0]),
                        leaves,
                    )
                )
            )
        )
        common_seg = np.array(img.vs["name"], dtype=int)[common_seg]
        fork_t = adata.uns["graph"]["pp_info"].loc[common_seg, "time"].max()

        g_early = []
        dfs = []
        for i in range(len(leaves)):
            included = pd.concat([s[i] for s in stats], axis=1).mean(axis=1)
            df = included[~np.isnan(included)].sort_values()
            adata.uns[name]["fork"].loc[df.index, "inclusion"] = df.values
            g_early = g_early + [df.index[df < fork_t]]
            dfs = dfs + [df]

        if identify_early_features:
            updated = " and 'module'."
            adata.uns[name]["fork"]["module"] = np.nan
            for df, g_e in zip(dfs, g_early):
                adata.uns[name]["fork"].loc[df.index, "module"] = "late"
                adata.uns[name]["fork"].loc[g_e, "module"] = "early"

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "added \n"
        "    .uns['"
        + name
        + "']['module_inclusion"
        + perm_str
        + "'], milestone specific dataframes containing inclusion timing for each gene in each probabilistic cells projection.\n"
        + "    .uns['"
        + name
        + "']['fork'] has been updated with the column 'inclusion'"
        + updated
    )

    return adata if copy else None



"""Settings
"""
from . import logging
import matplotlib.pyplot as plt

verbosity = 3
"""Verbosity level (0=errors, 1=warnings, 2=info, 3=hints)
"""

logfile = ""
"""Name of logfile. By default is set to '' and writes to standard output."""

plot_prefix = "scFates_"
"""Global prefix that is appended to figure filenames.
"""

plot_suffix = ""
"""Global suffix that is appended to figure filenames.
"""


def _set_start_time():
    from time import time

    return time()


_start = _set_start_time()
"""Time when the settings module is first imported."""

_previous_time = _start
"""Variable for timing program parts."""

_previous_memory_usage = -1
"""Stores the previous memory usage."""


def set_figure_pubready():
    plt.rcParams["svg.fonttype"] = "none"
    plt.rcParams["font.family"] = ["sans-serif"]
    plt.rcParams["font.sans-serif"] = ["Arial"]
    plt.rcParams["pdf.fonttype"] = 42
    plt.rcParams["savefig.bbox"] = "tight"



import os

os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"

from typing import Union, Optional, Iterable

import numpy as np
import pandas as pd
from functools import partial
from anndata import AnnData
import shutil
import sys
import copy
from statsmodels.stats.multitest import multipletests
import igraph
import warnings


from joblib import delayed
from tqdm import tqdm

from .. import logging as logg
from .. import settings
from ..plot.test_association import test_association as plot_test_association
from .utils import getpath, get_X, ProgressParallel, importeR

Rpy2, R, rstats, rmgcv, Formula = importeR("testing feature association to the tree")
check = [type(imp) == str for imp in [Rpy2, R, rstats, rmgcv, Formula]]


def test_association(
    adata: AnnData,
    n_map: int = 1,
    n_jobs: int = 1,
    spline_df: int = 5,
    fdr_cut: float = 1e-4,
    A_cut: int = 1,
    st_cut: float = 0.8,
    reapply_filters: bool = False,
    plot: bool = False,
    root=None,
    leaves=None,
    copy: bool = False,
    layer: Optional[str] = None,
):

    """\
    Determine a set of genes significantly associated with the trajectory.


    Feature expression is modeled as a function of pseudotime in a branch-specific manner,
    using cubic spline regression :math:`g_{i} \\sim\ t_{i}` for each branch independently.
    This tree-dependent model is then compared with an unconstrained model :math:`g_{i} \\sim\ 1`
    using F-test.

    The models are fit using *mgcv* R package.

    Benjamini-Hochberg correction is used to adjust for multiple hypothesis testing.


    Parameters
    ----------
    adata
        Annotated data matrix.
    layer
        adata layer to use for the test.
    n_map
        number of cell mappings from which to do the test.
    n_jobs
        number of cpu processes used to perform the test.
    spline_df
        dimension of the basis used to represent the smooth term (has no effect for now).
    fdr_cut
        FDR (Benjamini-Hochberg adjustment) cutoff on significance; significance if FDR < fdr_cut.
    A_cut
        amplitude is max of predicted value minus min of predicted value by GAM. significance if A > A_cut.
    st_cut
        cutoff on stability (fraction of mappings with significant (fdr,A) pair) of association; significance, significance if st > st_cut.
    reapply_filters
        avoid recomputation and reapply fitlers.
    plot
        call scf.pl.test_associationa after the test.
    root
        restrain the test to a subset of the tree (in combination with leaves).
    leaves
        restrain the test to a subset of the tree (in combination with root).
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

        `.var['p_val']`
            p-values from statistical test.
        `.var['fdr']`
            corrected values from multiple testing.
        `.var['st']`
            proportion of mapping in which feature is significant.
        `.var['A']`
            amplitue of change of tested feature.
        '.var['signi']`
            feature is significantly changing along pseuodtime
        `.uns['stat_assoc_list']`
            list of fitted features on the tree for all mappings.
    """

    if any(check):
        idx = np.argwhere(
            [type(imp) == str for imp in [Rpy2, R, rstats, rmgcv, Formula]]
        ).min()
        raise Exception(np.array([Rpy2, R, rstats, rmgcv, Formula])[idx])

    adata = adata.copy() if copy else adata

    if "pseudotime_list" not in adata.uns:
        raise ValueError(
            "You need to run `tl.pseudotime` before testing for association."
        )

    graph = adata.uns["graph"]

    mlsc_temp = None
    if leaves is not None:
        # weird hack to keep milestones colors saved
        if "milestones_colors" in adata.uns:
            mlsc = adata.uns["milestones_colors"].copy()
            mlsc_temp = mlsc.copy()

        dct = graph["milestones"]
        keys = np.array(list(dct.keys()))
        vals = np.array(list(dct.values()))

        leaves = list(map(lambda leave: dct[leave], leaves))
        root = dct[root]

    if reapply_filters & ("stat_assoc_list" in adata.uns):
        stat_assoc_l = list(adata.uns["stat_assoc_list"].values())
        # stat_assoc_l = list(map(lambda x: pd.DataFrame(x,index=x["features"]),stat_assoc_l))
        adata = apply_filters(adata, stat_assoc_l, fdr_cut, A_cut, st_cut)

        logg.info(
            "reapplied filters, "
            + str(sum(adata.var["signi"]))
            + " significant features"
        )

        if plot:
            plot_test_association(adata)

        return adata if copy else None

    genes = adata.var_names
    if root is None:
        cells = adata.obs_names
    else:
        df = adata.obs.copy()
        edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
        img = igraph.Graph()
        img.add_vertices(
            np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
        )
        img.add_edges(edges)

        cells = np.unique(
            np.concatenate(
                list(
                    map(
                        lambda leave: getpath(
                            img, root, graph["tips"], leave, graph, df
                        ).index,
                        leaves,
                    )
                )
            )
        )

    Xgenes = get_X(adata, cells, genes, layer, togenelist=True)

    logg.info("test features for association with the trajectory", reset=True, end="\n")

    stat_assoc_l = list()

    def test_assoc_map(m):
        data = list(
            zip(
                [adata.uns["pseudotime_list"][str(m)].loc[cells, :]] * len(Xgenes),
                Xgenes,
            )
        )

        stat = ProgressParallel(
            n_jobs=n_jobs if n_map == 1 else 1,
            total=len(data),
            file=sys.stdout,
            use_tqdm=n_map == 1,
            desc="    single mapping ",
        )(delayed(gt_fun_exp)(data[d]) for d in range(len(data)))
        stat = pd.DataFrame(stat, index=genes, columns=["p_val", "A"])
        stat["fdr"] = multipletests(stat.p_val, method="bonferroni")[1]
        return stat

    stat_assoc_l = ProgressParallel(
        n_jobs=1 if n_map == 1 else n_jobs,
        total=n_map,
        file=sys.stdout,
        use_tqdm=n_map > 1,
        desc="    multi mapping ",
    )(delayed(test_assoc_map)(m) for m in range(n_map))
    # stat_assoc_l = stat_assoc_l + [stat]

    adata = apply_filters(adata, stat_assoc_l, fdr_cut, A_cut, st_cut)

    if mlsc_temp is not None:
        adata.uns["milestones_colors"] = mlsc_temp

    logg.info(
        "    found " + str(sum(adata.var["signi"])) + " significant features",
        time=True,
        end=" " if settings.verbosity > 2 else "\n",
    )
    logg.hint(
        "added\n"
        "    .var['p_val'] values from statistical test.\n"
        "    .var['fdr'] corrected values from multiple testing.\n"
        "    .var['st'] proportion of mapping in which feature is significant.\n"
        "    .var['A'] amplitue of change of tested feature.\n"
        "    .var['signi'] feature is significantly changing along pseudotime.\n"
        "    .uns['stat_assoc_list'] list of fitted features on the graph for all mappings."
    )

    if plot:
        plot_test_association(adata)

    return adata if copy else None


def gt_fun_exp(data):
    sdf = data[0]
    sdf["exp"] = data[1]

    global rmgcv
    global rstats

    def gamfit(s):
        m = rmgcv.gam(Formula("exp~s(t,k=5)"), data=sdf.loc[sdf["seg"] == s, :])
        return dict({"d": m[5][0], "df": m[42][0], "p": rmgcv.predict_gam(m)})

    mdl = list(map(gamfit, sdf.seg.unique()))
    mdf = pd.concat(list(map(lambda x: pd.DataFrame([x["d"], x["df"]]), mdl)), axis=1).T
    mdf.columns = ["d", "df"]

    odf = sum(mdf["df"]) - mdf.shape[0]
    m0 = rmgcv.gam(Formula("exp~1"), data=sdf)
    if sum(mdf["d"]) == 0:
        fstat = 0
    else:
        fstat = (m0[5][0] - sum(mdf["d"])) / (m0[42][0] - odf) / (sum(mdf["d"]) / odf)

    df_res0 = m0[42][0]
    df_res_odf = df_res0 - odf
    pval = rstats.pf(fstat, df_res_odf, odf, lower_tail=False)[0]
    pr = np.concatenate(list(map(lambda x: x["p"], mdl)))

    return [pval, max(pr) - min(pr)]


def apply_filters(adata, stat_assoc_l, fdr_cut, A_cut, st_cut, prefix=""):
    n_map = len(stat_assoc_l)
    if n_map > 1:
        stat_assoc = pd.DataFrame(
            {
                prefix
                + "p_val": pd.concat(
                    list(map(lambda x: x[prefix + "p_val"], stat_assoc_l)), axis=1
                ).median(axis=1),
                prefix
                + "A": pd.concat(
                    list(map(lambda x: x[prefix + "A"], stat_assoc_l)), axis=1
                ).median(axis=1),
                prefix
                + "fdr": pd.concat(
                    list(map(lambda x: x[prefix + "fdr"], stat_assoc_l)), axis=1
                ).median(axis=1),
                prefix
                + "st": pd.concat(
                    list(
                        map(
                            lambda x: (x[prefix + "fdr"] < fdr_cut)
                            & (x[prefix + "A"] > A_cut),
                            stat_assoc_l,
                        )
                    ),
                    axis=1,
                ).sum(axis=1)
                / n_map,
            }
        )
    else:
        stat_assoc = stat_assoc_l[0]
        stat_assoc[prefix + "st"] = (
            (stat_assoc[prefix + "fdr"] < fdr_cut) & (stat_assoc[prefix + "A"] > A_cut)
        ) * 1

    # saving results
    stat_assoc[prefix + "signi"] = stat_assoc[prefix + "st"] > st_cut

    if set(stat_assoc.columns.tolist()).issubset(adata.var.columns):
        adata.var[stat_assoc.columns] = stat_assoc
    else:
        adata.var = pd.concat([adata.var, stat_assoc], axis=1)

    names = np.arange(len(stat_assoc_l)).astype(str).tolist()

    dictionary = dict(zip(names, stat_assoc_l))
    adata.uns[prefix + "stat_assoc_list"] = dictionary

    return adata



from typing import Union, Sequence
import numpy as np
from matplotlib import pyplot as pl
from cycler import Cycler, cycler
from matplotlib import rcParams
from . import palettes
from matplotlib.colors import to_hex


def _set_colors_for_categorical_obs(
    adata,
    value_to_plot,
    palette: Union[str, Sequence[str], Cycler],
):
    """
    Sets the adata.uns[value_to_plot + '_colors'] according to the given palette
    Parameters
    ----------
    adata
        annData object
    value_to_plot
        name of a valid categorical observation
    palette
        Palette should be either a valid :func:`~matplotlib.pyplot.colormaps` string,
        a sequence of colors (in a format that can be understood by matplotlib,
        eg. RGB, RGBS, hex, or a cycler object with key='color'
    Returns
    -------
    None
    """

    categories = adata.obs[value_to_plot].cat.categories
    # check is palette is a valid matplotlib colormap
    if isinstance(palette, str) and palette in pl.colormaps():
        # this creates a palette from a colormap. E.g. 'Accent, Dark2, tab20'
        cmap = pl.get_cmap(palette)
        colors_list = [to_hex(x) for x in cmap(np.linspace(0, 1, len(categories)))]

    else:
        # check if palette is a list and convert it to a cycler, thus
        # it doesnt matter if the list is shorter than the categories length:
        if isinstance(palette, cabc.Sequence):
            if len(palette) < len(categories):
                logg.warning(
                    "Length of palette colors is smaller than the number of "
                    f"categories (palette length: {len(palette)}, "
                    f"categories length: {len(categories)}. "
                    "Some categories will have the same color."
                )
            # check that colors are valid
            _color_list = []
            for color in palette:
                if not is_color_like(color):
                    # check if the color is a valid R color and translate it
                    # to a valid hex color value
                    if color in additional_colors:
                        color = additional_colors[color]
                    else:
                        raise ValueError(
                            "The following color value of the given palette "
                            f"is not valid: {color}"
                        )
                _color_list.append(color)

            palette = cycler(color=_color_list)
        if not isinstance(palette, Cycler):
            raise ValueError(
                "Please check that the value of 'palette' is a valid "
                "matplotlib colormap string (eg. Set2), a  list of color names "
                "or a cycler with a 'color' key."
            )
        if "color" not in palette.keys:
            raise ValueError("Please set the palette key 'color'.")

        cc = palette()
        colors_list = [to_hex(next(cc)["color"]) for x in range(len(categories))]

    adata.uns[value_to_plot + "_colors"] = colors_list


def _set_default_colors_for_categorical_obs(adata, value_to_plot):
    """
    Sets the adata.uns[value_to_plot + '_colors'] using default color palettes
    Parameters
    ----------
    adata
        AnnData object
    value_to_plot
        Name of a valid categorical observation
    Returns
    -------
    None
    """

    categories = adata.obs[value_to_plot].cat.categories
    length = len(categories)

    # check if default matplotlib palette has enough colors
    if len(rcParams["axes.prop_cycle"].by_key()["color"]) >= length:
        cc = rcParams["axes.prop_cycle"]()
        palette = [next(cc)["color"] for _ in range(length)]
        palette = [to_hex(x) for x in palette]
    else:
        if length <= 20:
            palette = palettes.default_20
        elif length <= 28:
            palette = palettes.default_28
        elif length <= len(palettes.default_102):  # 103 colors
            palette = palettes.default_102
        else:
            palette = ["grey" for _ in range(length)]
            logg.info(
                f"the obs value {value_to_plot!r} has more than 103 categories. Uniform "
                "'grey' color will be used for all categories."
            )

    adata.uns[value_to_plot + "_colors"] = palette[:length]


def add_colors_for_categorical_sample_annotation(
    adata, key, palette=None, force_update_colors=False
):

    color_key = f"{key}_colors"
    colors_needed = len(adata.obs[key].cat.categories)
    if palette and force_update_colors:
        _set_colors_for_categorical_obs(adata, key, palette)
    elif color_key in adata.uns and len(adata.uns[color_key]) <= colors_needed:
        _validate_palette(adata, key)
    else:
        _set_default_colors_for_categorical_obs(adata, key)



import os
from os import path
from setuptools import setup, find_packages

dir = path.abspath(path.dirname(__file__))
with open("requirements.txt") as f:
    requirements = f.read().splitlines()

with open(path.join(dir, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


on_rtd = os.environ.get("READTHEDOCS") == "True"
if on_rtd:
    requirements = []


def set_version():
    head = open(path.join(dir, ".git", "HEAD")).readline()
    if head.find("dev") != -1:
        return {
            "template": "{tag}.dev{ccount}",
            "dev_template": "{tag}.dev{ccount}",
            "dirty_template": "{tag}.dev{ccount}",
        }
    else:
        return {"template": "{tag}", "dev_template": "{tag}", "dirty_template": "{tag}"}


setup(
    name="scFates",
    version_config=set_version(),
    setup_requires=["setuptools-git-versioning"],
    description="scanpy compatible python suite for fast tree inference and advanced pseudotime downstream analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/LouisFaure/scFates",
    author="Louis Faure",
    author_email="",
    packages=find_packages(),
    package_dir={"scFates": "scFates"},
    install_requires=requirements,
    include_package_data=True,
    package_data={"": ["datasets/*.h5ad"]},
    zip_safe=False,
)



import numpy as np
import pandas as pd
from anndata import AnnData

import igraph
import seaborn as sns
import matplotlib.pyplot as plt
from scipy import stats
from adjustText import adjust_text
from matplotlib import patches
from scipy import sparse
import matplotlib.gridspec as gridspec

from typing import Union, Optional, List
from typing_extensions import Literal
from scanpy.plotting._utils import savefig_or_show
import scanpy as sc

from .. import logging as logg
from ..tools.utils import getpath, importeR
from .trajectory import trajectory
from .dendrogram import dendrogram

Rpy2, R, rstats, rmgcv, Formula = importeR("fitting associated features")
check = [type(imp) == str for imp in [Rpy2, R, rstats, rmgcv, Formula]]

from ..get import modules as get_modules
from .trajectory import remove_info
from .utils import gen_milestones_gradients
from .. import logging as logg


def trends(
    adata: AnnData,
    features=None,
    cluster=None,
    highlight_features: Union[List, Literal["A", "fdr"]] = "A",
    n_features: int = 10,
    root_milestone: Union[None, str] = None,
    milestones: Union[None, str] = None,
    module: Union[None, Literal["early", "late"]] = None,
    branch: Union[None, str] = None,
    annot: Union[None, str] = None,
    title: str = "",
    feature_cmap: str = "RdBu_r",
    pseudo_cmap: str = "viridis",
    plot_emb: bool = True,
    plot_heatmap: bool = True,
    wspace: Union[None, float] = None,
    show_segs: bool = True,
    basis: str = "umap",
    heatmap_space: float = 0.5,
    offset_names: float = 0.15,
    fontsize: int = 9,
    style: Literal["normal", "italic", "oblique"] = "normal",
    ordering: Union[
        None, Literal["pearson", "spearman", "quantile", "max"]
    ] = "pearson",
    ord_thre=0.7,
    filter_complex: bool = False,
    complex_thre: float = 0.7,
    complex_z: float = 3,
    figsize: Union[None, tuple] = None,
    axemb=None,
    show: Optional[bool] = None,
    output_mean: bool = False,
    save: Union[str, bool, None] = None,
    save_genes: Optional[bool] = None,
    **kwargs,
):

    """\
    Plot a set of fitted features over pseudotime.

    Parameters
    ----------
    adata
        Annotated data matrix.
    features
        Name of the fitted features.
    highlight_features
        which features will be annotated on the heatmap, by default, features with highest amplitude are shown.
    n_features
        number of top features to show if no list are provided.
    root_milestone
        tip defining progenitor branch.
    milestones
        tips defining the progenies branches.
    module
        if bifurcation analysis as been performed, subset features to a specific module.
    branch
        if bifurcation analysis as been performed, subset features to a specific milestone.
    annot
        adds an annotation row on top of the heatmap.
    title
        add a title.
    feature_cmap
        colormap for features.
    pseudo_cmap
        colormap for pseudotime.
    plot_emb
        call pl.trajectory on the left side.
    plot_heatmap
        show heatmap on the right side.
    wspace
        width space between emb and heatmap.
    show_segs
        display segments on emb.
    basis
        Name of the `obsm` basis to use if plot_emb is True.
    heatmap_space
        how much space does the heatmap take, in proportion of the whole plot space.
    offset_names
        how far on the right the annotated features should be displayed, in proportion of the heatmap space.
    fontsize
        font size of the feature annotations.
    style
        font style.
    ordering
        strategy to order the features on heatmap, quantile takes the mean pseudotime of the choosen value.
    ord_thre
        for 'max': proportion of maximum of fitted value to consider to compute the mean pseudotime.
        for 'quantile': quantile to consider to compute the mean pseudotime.
        for 'pearson'/'spearman': proportion of max value to assign starting cell.
    filter_complex
        filter out complex features
    complex_thre
        quantile to consider for complex filtering
    complex_z
        standard deviation threshold on the quantile subsetted feature.
    figsize
        figure size.
    axemb
        existing ax for plotting emb
    output_mean
        output mean fitted values to adata.
    show
        show the plot.
    save
        save the plot.
    save_genes
        save list of genes following the order displayed on the heatmap.
    **kwargs
        if `plot_emb=True`, arguments passed to :func:`scFates.pl.trajectory` or :func:`scFates.pl.dendrogram` if `basis="dendro"`

    Returns
    -------
    If `show==False` a matrix of :class:`~matplotlib.axes.Axes`

    """

    offset_heatmap = 1 - heatmap_space

    if plot_emb:
        adata_temp = adata.copy()

    graph = adata.uns["graph"]

    if milestones is not None:
        adata = adata.copy()
        dct = graph["milestones"]

        leaves = list(map(lambda leave: dct[leave], milestones))
        root = dct[root_milestone]
        df = adata.obs.copy(deep=True)
        edges = (
            adata.uns["graph"]["pp_seg"][["from", "to"]]
            .astype(str)
            .apply(tuple, axis=1)
            .values
        )
        img = igraph.Graph()
        img.add_vertices(
            np.unique(
                adata.uns["graph"]["pp_seg"][["from", "to"]]
                .values.flatten()
                .astype(str)
            )
        )
        img.add_edges(edges)

        cells = np.unique(
            np.concatenate(
                list(
                    map(
                        lambda leave: getpath(
                            img,
                            root,
                            adata.uns["graph"]["tips"],
                            leave,
                            adata.uns["graph"],
                            df,
                        ).index,
                        leaves,
                    )
                )
            )
        )

        adata = adata[cells]

    if (features is None) & (cluster is None):
        features = adata.var_names
    if cluster is not None:
        features = adata.var_names[adata.var.fit_clusters == cluster]

    if branch is not None:
        name = root_milestone + "->" + "<>".join(milestones)
        df = adata.uns[name]["fork"]
        if module is not None:
            sel = (df.branch == branch) & (df.module == module)
        else:
            sel = df.branch == branch
        features = df.loc[sel, :].index

    fitted = pd.DataFrame(
        adata.layers["fitted"], index=adata.obs_names, columns=adata.var_names
    ).T.copy(deep=True)
    g = adata.obs.groupby("seg")
    seg_order = g.apply(lambda x: np.min(x.t)).sort_values().index.tolist()
    cell_order = np.concatenate(
        list(
            map(
                lambda x: adata.obs.t[adata.obs.seg == x].sort_values().index, seg_order
            )
        )
    )
    fitted = fitted.loc[:, cell_order]
    # fitted=fitted.apply(lambda x: (x-x.mean())/x.std(),axis=1)

    fitted = fitted.apply(lambda x: (x - x.min()) / (x.max() - x.min()), axis=1)

    if filter_complex:
        # remove complex features using quantiles
        varia = list(
            map(
                lambda x: adata.obs.t[cell_order][
                    fitted.loc[x, :].values
                    > np.quantile(fitted.loc[x, :].values, q=complex_thre)
                ].var(),
                features,
            )
        )
        z = np.abs(stats.zscore(varia))
        torem = np.argwhere(z > complex_z).flatten()

        if len(torem) > 0:
            logg.info("found " + str(len(torem)) + " complex fitted features")
            logg.hint("added\n" + "    'complex' column in (adata.var)")
            adata.var["complex"] = False
            # adata.var.iloc[torem,"complex"]=True
            adata.var.loc[fitted.index[torem], "complex"] = True
            features = adata.var_names[~adata.var["complex"]]

    fitted = fitted.loc[features, :]

    if ordering == "quantile":
        feature_order = (
            fitted.apply(
                lambda x: adata.obs.t[fitted.columns][
                    x > np.quantile(x, q=ord_thre)
                ].mean(),
                axis=1,
            )
            .sort_values()
            .index
        )
    elif ordering == "max":
        feature_order = (
            fitted.apply(
                lambda x: adata.obs.t[fitted.columns][
                    (x - x.min()) / (x.max() - x.min()) > ord_thre
                ].mean(),
                axis=1,
            )
            .sort_values()
            .index
        )
    elif ordering in ("pearson", "spearman"):
        start_feature = (
            fitted.apply(
                lambda x: adata.obs.t[fitted.columns][
                    (x - x.min()) / (x.max() - x.min()) > ord_thre
                ].mean(),
                axis=1,
            )
            .sort_values()
            .index[0]
        )
        feature_order = (
            fitted.T.corr(method=ordering)
            .loc[start_feature, :]
            .sort_values(ascending=False)
            .index
        )

    else:
        feature_order = fitted.index

    fitted_sorted = fitted.loc[feature_order, :]

    if annot == "milestones":
        annot_cmap = gen_milestones_gradients(adata, seg_order)

    elif type(annot) == str:
        if len(adata.obs[annot].unique()) > 1:
            color_key = annot + "_colors"
            if color_key not in adata.uns or len(adata.uns[color_key]) == 1:
                from . import palette_tools

                palette_tools._set_default_colors_for_categorical_obs(adata, annot)
            annot_cmap = pd.Series(
                list(
                    map(
                        lambda c: adata.uns[color_key][
                            adata.obs[annot].cat.categories == c
                        ][0],
                        adata.obs[annot][fitted_sorted.columns].values,
                    )
                ),
                index=fitted_sorted.columns,
            )
    else:
        annot = None

    if axemb is None:
        ratio = plt.rcParams["figure.figsize"][0] / plt.rcParams["figure.figsize"][1]
        figsize = (
            ((4 * plot_emb * ratio) + 4 * plot_heatmap, 4)
            if figsize is None
            else figsize
        )
        fig = plt.figure(figsize=figsize)
        gsubs = gridspec.GridSpec(
            1,
            2,
            figure=fig,
            width_ratios=[1 * plot_emb * ratio, 1 * plot_heatmap],
            wspace=wspace,
        )
        axs = []
    if plot_heatmap:
        gs_ht = gridspec.GridSpecFromSubplotSpec(
            2 + (annot is not None),
            1,
            height_ratios=(1, 1, 18) if (annot is not None) else (1, 19),
            subplot_spec=gsubs[1],
            hspace=0,
        )
        i = 0
        if annot is not None:
            axannot = plt.subplot(gs_ht[i])
            axs = axs + [axannot]
            sns.heatmap(
                pd.DataFrame(range(fitted_sorted.shape[1])).T,
                robust=False,
                rasterized=True,
                cmap=annot_cmap[fitted_sorted.columns].values.tolist(),
                xticklabels=False,
                yticklabels=False,
                cbar=False,
                ax=axannot,
            )
            i = i + 1

        axpsdt = plt.subplot(gs_ht[i])
        axs = axs + [axpsdt]
        sns.heatmap(
            pd.DataFrame(adata.obs.t[fitted_sorted.columns].values).T,
            robust=True,
            rasterized=True,
            cmap=pseudo_cmap,
            xticklabels=False,
            yticklabels=False,
            cbar=False,
            ax=axpsdt,
            vmax=adata.obs.t.max(),
        )

        axheatmap = plt.subplot(gs_ht[i + 1])
        axs = axs + [axheatmap]
        sns.heatmap(
            fitted_sorted,
            robust=True,
            cmap=feature_cmap,
            rasterized=True,
            xticklabels=False,
            yticklabels=False,
            ax=axheatmap,
            cbar=False,
        )

        def add_frames(axis, vert):
            rect = patches.Rectangle(
                (0, 0),
                len(fitted_sorted.columns),
                vert,
                linewidth=1,
                edgecolor="k",
                facecolor="none",
            )
            # Add the patch to the Axes
            axis.add_patch(rect)
            offset = 0
            for s in seg_order[:-1]:
                prev_offset = offset
                offset = offset + (adata.obs.seg == s).sum()
                rect = patches.Rectangle(
                    (prev_offset, 0),
                    (adata.obs.seg == s).sum(),
                    vert,
                    linewidth=1,
                    edgecolor="k",
                    facecolor="none",
                )
                axis.add_patch(rect)
            return axis

        axpsdt = add_frames(axpsdt, 1)

        if annot is not None:
            axannot = add_frames(axannot, 1)

        axheatmap = add_frames(axheatmap, fitted_sorted.shape[0])

        if highlight_features == "A":
            highlight_features = (
                adata.var.A[features].sort_values(ascending=False)[:n_features].index
            )
        elif highlight_features == "fdr":
            highlight_features = (
                adata.var.fdr[features].sort_values(ascending=True)[:n_features].index
            )
        xs = np.repeat(fitted_sorted.shape[1], len(highlight_features))
        ys = (
            np.array(
                list(
                    map(
                        lambda g: np.argwhere(fitted_sorted.index == g)[0][0],
                        highlight_features,
                    )
                )
            )
            + 0.5
        )

        texts = []
        for x, y, s in zip(xs, ys, highlight_features):
            texts.append(axheatmap.text(x, y, s, fontsize=fontsize, style=style))

        patch = patches.Rectangle(
            (0, 0),
            fitted_sorted.shape[1] + fitted_sorted.shape[1] * offset_names,
            fitted_sorted.shape[0],
            alpha=0,
        )  # We add a rectangle to make sure the labels don't move to the right
        axpsdt.set_xlim(
            (0, fitted_sorted.shape[1] + fitted_sorted.shape[1] * offset_heatmap)
        )
        if annot is not None:
            axannot.set_xlim(
                (0, fitted_sorted.shape[1] + fitted_sorted.shape[1] * offset_heatmap)
            )
        axheatmap.set_xlim(
            (0, fitted_sorted.shape[1] + fitted_sorted.shape[1] * offset_heatmap)
        )
        axheatmap.add_patch(patch)
        axheatmap.hlines(
            fitted_sorted.shape[0], 0, fitted_sorted.shape[1], color="k", clip_on=True
        )

        adjust_text(
            texts,
            ax=axheatmap,
            add_objects=[patch],
            va="center",
            ha="left",
            autoalign=False,
            expand_text=(1.05, 1.2),
            lim=5000,
            only_move={"text": "y", "objects": "x"},
            precision=0.1,
            expand_points=(1.2, 1.05),
        )

        for i in range(len(xs)):
            xx = [
                xs[i] + 1,
                fitted_sorted.shape[1] + fitted_sorted.shape[1] * offset_names,
            ]
            yy = [ys[i], texts[i].get_position()[1]]
            axheatmap.plot(xx, yy, color="k", linewidth=0.75)

    if plot_emb:
        if axemb is None:
            gs_emb = gridspec.GridSpecFromSubplotSpec(1, 1, subplot_spec=gsubs[0])
            axemb = fig.add_subplot(gs_emb[0])
            axs = axs + [axemb]
        adata_temp.obs["mean_trajectory"] = 0
        adata_temp.obs.loc[
            fitted_sorted.columns, "mean_trajectory"
        ] = fitted_sorted.mean(axis=0).values
        if output_mean:
            adata.obs["mean_trajectory"] = adata_temp.obs["mean_trajectory"]
        if "color" in kwargs:
            color = kwargs["color"]
            del kwargs["color"]
        else:
            color = "mean_trajectory"

        if "cmap" in kwargs:
            feature_cmap = kwargs["cmap"]
            del kwargs["cmap"]

        if basis == "dendro":
            dendrogram(
                adata_temp,
                color=color,
                cmap=feature_cmap,
                ax=axemb,
                title=title,
                root_milestone=root_milestone,
                milestones=milestones,
                show_info=False,
                **kwargs,
            )
        elif show_segs == False:
            sc.pl.embedding(
                adata=adata_temp,
                basis=basis,
                color=color,
                cmap=feature_cmap,
                ax=axemb,
                title=title,
                show=False,
                **kwargs,
            )
            remove_info(adata_temp, axemb, color)
        else:
            trajectory(
                adata=adata_temp,
                basis=basis,
                color_seg=color,
                color_cells=color,
                cmap_seg=feature_cmap,
                cmap=feature_cmap,
                show_info=False,
                ax=axemb,
                title=title,
                root_milestone=root_milestone,
                milestones=milestones,
                show=False,
                **kwargs,
            )

    if save_genes is not None:
        with open(save_genes, "w") as f:
            for item in fitted_sorted.index:
                f.write("%s\n" % item)

    if show == False:
        return axs if plot_heatmap else axemb
    if save is not None:
        savefile = "figures/trends" + save
        logg.warn("saving figure to file " + savefile)
        fig.savefig(savefile, bbox_inches="tight")


def single_trend(
    adata: AnnData,
    feature: Union[str, None] = None,
    root_milestone: Union[None, str] = None,
    milestones: Union[None, str] = None,
    module: Union[None, Literal["early", "late"]] = None,
    branch: Union[None, str] = None,
    basis: str = "umap",
    ylab: str = "expression",
    color_exp=None,
    alpha_expr: float = 0.3,
    size_expr: float = 2,
    fitted_linewidth: float = 2,
    layer: Union[str, None] = None,
    cmap_seg: str = "RdBu_r",
    cmap_cells: str = "RdBu_r",
    plot_emb: bool = True,
    wspace: Union[None, float] = None,
    figsize: tuple = (8, 4),
    ax_trend=None,
    ax_emb=None,
    show: Optional[bool] = None,
    save: Union[str, bool, None] = None,
    **kwargs,
):

    """\
    Plot a single feature fit over pseudotime.

    Parameters
    ----------
    adata
        Annotated data matrix.
    feature
        Name of the fitted feature.
    root_milestone
        if plotting module instead of feature, tip defining progenitor branch.
    milestones
        if plotting module instead of feature, tips defining the progenies branches.
    module
        if plotting module instead of feature, whether to plot early or late modules.
    branch
        if plotting module instead of feature, plot fitted milestone-specific module.
    basis
        Name of the `obsm` basis to use.
    ylab
        ylabel of right plot.
    colo_rexp
        color of raw datapoints on right plot.
    alpha_expr
        alpha of raw datapoints on right plot.
    size_expr
        size of raw datapoints on right plot.
    fitted_linewidth
        linewidth of fitted line on right plot.
    layer
        layer to plot for the raw datapoints.
    cmap_seg
        colormap for trajectory segments on left plot.
    cmap_cells
        colormap for cells on left plot.
    plot_emb
        plot the emb on the left side.
    wspace
        width space between emb and heatmap.
    figsize
        figure size in inches.
    ax_trend
        existing ax for trends, only works when emb plot is disabled.
    ax_emb
        existing ax for embedding plot.
    show
        show the plot.
    save
        save the plot.

    Returns
    -------
    If `show==False` a tuple of two :class:`~matplotlib.axes.Axes`

    """
    if root_milestone is None:
        color_key = "seg_colors"
        if color_key not in adata.uns:
            from . import palette_tools

            palette_tools._set_default_colors_for_categorical_obs(adata, "seg")

        if layer is None:
            if sparse.issparse(adata.X):
                Xfeature = adata[:, feature].X.A.T.flatten()
            else:
                Xfeature = adata[:, feature].X.T.flatten()
        else:
            if sparse.issparse(adata.layers[layer]):
                Xfeature = adata[:, feature].layers[layer].A.T.flatten()
            else:
                Xfeature = adata[:, feature].layers[layer].T.flatten()

        df = pd.DataFrame(
            {
                "t": adata.obs.t,
                "fitted": adata[:, feature].layers["fitted"].flatten(),
                "expression": Xfeature,
                "seg": adata.obs.seg,
            }
        ).sort_values("t")
    else:
        if plot_emb:
            logg.warn("trajectory plotting is not yet implemented for module plot!")
            plot_emb = False
        graph = adata.uns["graph"]
        miles_cat = adata.obs.milestones.cat.categories
        mil_col = np.array(adata.uns["milestones_colors"])

        X_modules = get_modules(adata, root_milestone, milestones, layer)
        X_early, X_late = X_modules.iloc[:, :2], X_modules.iloc[:, 2:]

        X = pd.concat([X_early, X_late], axis=1)
        X["t"] = adata.obs.loc[X.index, "t"].values

        def gamfit(f):
            m = rmgcv.gam(
                Formula(f),
                data=X,
                gamma=1.5,
            )
            return pd.Series(rmgcv.predict_gam(m), index=X.index)

        predictions = gamfit("{}_{}~s(t,bs='ts')".format(module, branch))

        df = pd.DataFrame(
            {
                "t": adata[X.index].obs.t,
                "fitted": predictions.values,
                "expression": X["_".join([module, branch])].values,
                "seg": adata[X.index].obs.seg,
            }
        ).sort_values("t")

        color_exp = mil_col[miles_cat == branch][0]
    ratio = plt.rcParams["figure.figsize"][0] / plt.rcParams["figure.figsize"][1]
    if (ax_emb is None) & (ax_trend is None):
        if plot_emb:
            fig, (ax_emb, ax_trend) = plt.subplots(
                1,
                2,
                figsize=figsize,
                gridspec_kw=dict(width_ratios=[1 * ratio, 1], wspace=wspace),
            )
        else:
            fig, ax_trend = plt.subplots(1, 1, figsize=figsize)
            # axs = ["empty", axs]

    for s in df.seg.unique():
        if color_exp is None:
            ax_trend.scatter(
                df.loc[df.seg == s, "t"],
                df.loc[df.seg == s, "expression"],
                alpha=alpha_expr,
                s=size_expr,
                c=adata.uns["seg_colors"][
                    np.argwhere(adata.obs.seg.cat.categories == s)[0][0]
                ],
            )
            ax_trend.plot(
                df.loc[df.seg == s, "t"],
                df.loc[df.seg == s, "fitted"],
                c=adata.uns["seg_colors"][
                    np.argwhere(adata.obs.seg.cat.categories == s)[0][0]
                ],
                linewidth=fitted_linewidth,
            )
            tolink = adata.uns["graph"]["pp_seg"].loc[int(s), "to"]
            for next_s in adata.uns["graph"]["pp_seg"].n.iloc[
                np.argwhere(
                    adata.uns["graph"]["pp_seg"].loc[:, "from"].isin([tolink]).values
                ).flatten()
            ]:
                ax_trend.plot(
                    [
                        df.loc[df.seg == s, "t"].iloc[-1],
                        df.loc[df.seg == next_s, "t"].iloc[0],
                    ],
                    [
                        df.loc[df.seg == s, "fitted"].iloc[-1],
                        df.loc[df.seg == next_s, "fitted"].iloc[0],
                    ],
                    c=adata.uns["seg_colors"][
                        np.argwhere(adata.obs.seg.cat.categories == next_s)[0][0]
                    ],
                    linewidth=fitted_linewidth,
                )
        else:
            ax_trend.scatter(
                df.loc[df.seg == s, "t"],
                df.loc[df.seg == s, "expression"],
                c=color_exp,
                alpha=alpha_expr,
                s=size_expr,
            )
            ax_trend.plot(
                df.loc[df.seg == s, "t"],
                df.loc[df.seg == s, "fitted"],
                c=color_exp,
                linewidth=fitted_linewidth,
            )
            tolink = adata.uns["graph"]["pp_seg"].loc[int(s), "to"]
            for next_s in adata.uns["graph"]["pp_seg"].n.iloc[
                np.argwhere(
                    adata.uns["graph"]["pp_seg"].loc[:, "from"].isin([tolink]).values
                ).flatten()
            ]:
                ax_trend.plot(
                    [
                        df.loc[df.seg == s, "t"].iloc[-1],
                        df.loc[df.seg == next_s, "t"].iloc[0],
                    ],
                    [
                        df.loc[df.seg == s, "fitted"].iloc[-1],
                        df.loc[df.seg == next_s, "fitted"].iloc[0],
                    ],
                    c=color_exp,
                    linewidth=fitted_linewidth,
                )

    ax_trend.set_ylabel(ylab)
    ax_trend.set_xlabel("pseudotime")
    x0, x1 = ax_trend.get_xlim()
    y0, y1 = ax_trend.get_ylim()
    if plot_emb:
        ax_trend.set_aspect(abs(x1 - x0) / abs(y1 - y0))

    if plot_emb:
        if basis == "dendro":
            dendrogram(
                adata,
                color=feature,
                cmap=cmap_cells,
                ax=ax_emb,
                title=feature,
                show_info=False,
                layer="fitted",
                show=False,
                save=False,
                **kwargs,
            )
        else:
            trajectory(
                adata,
                basis=basis,
                color_seg=feature,
                cmap_seg=cmap_seg,
                color_cells=feature,
                cmap=cmap_cells,
                show_info=False,
                ax=ax_emb,
                title=feature,
                layer="fitted",
                show=False,
                save=False,
                **kwargs,
            )

    savefig_or_show("single_trend", show=show, save=save)

    if show is False:
        return (ax_emb, ax_trend) if plot_emb else ax_trend



import networkx as nx
import random
import warnings
import numpy as np
import pandas as pd
import igraph
import matplotlib.pyplot as plt
import sys

from .utils import ProgressParallel
from joblib import delayed

import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import warnings

__all__ = ["swarmplot"]


from .. import logging as logg
from .. import settings

import anndata


def dendrogram(adata: anndata.AnnData, crowdedness: float = 1, n_jobs: int = 1):

    """\
    Generate a  single-cell dendrogram embedding.

    This representation aims in simplifying and abstracting the view of the tree, it follows URD style represenation of the cells.

    Parameters
    ----------
    adata
        Annotated data matrix.
    crowdedness
        will influence the repartition of the cells along the segments.

    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

        `.obs['X_dendro']`
           new embedding generated.
        `.uns['dendro_segments']`
            tree segments used for plotting.

    """

    logg.info("Generating dendrogram of tree", reset=True)

    roots = None

    graph = adata.uns["graph"]

    dct = graph["milestones"]
    keys = np.array(list(dct.keys()))
    vals = np.array(list(dct.values()))

    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
    img = igraph.Graph(directed=True)
    img.add_vertices(vals.astype(str))
    img.add_edges(edges)

    img.vs["label"] = keys

    pos = hierarchy_pos(img.to_networkx(), vert_gap=0.1)

    layout = pd.DataFrame(pos).T
    layout.index = np.array(list(graph["milestones"].values()), int)[
        np.array(list(pos.keys()))
    ]
    layout = layout[0]
    newseg = layout.loc[graph["pp_seg"].to]
    newseg.index = graph["pp_seg"].n
    df = adata.obs.copy()

    df["seg_pos"] = df.seg.astype(str)
    df["t"] = -df.t

    for i, x in enumerate(newseg):
        df.loc[df.seg == newseg.index[i], "seg_pos"] = x

    df.seg_pos = df.seg_pos.astype("category").cat.rename_categories(
        range(len(df.seg_pos.unique()))
    )
    df.seg_pos = df.seg_pos.astype(int).astype("category")

    dend = swarmplot(
        x="seg_pos", y="t", data=df, orient="v", size=crowdedness, n_jobs=n_jobs
    )
    adata.obsm["X_dendro"] = dend.loc[adata.obs_names].values

    # generate segments
    newseg = newseg[newseg.argsort()]

    segments = []
    for i, n in enumerate(newseg.index):
        fro = graph["pp_seg"].loc[int(n), "from"]
        to = graph["pp_seg"].loc[int(n), "to"]
        t_min = -graph["pp_info"].loc[fro].time
        t_max = -graph["pp_info"].loc[to].time

        segments = segments + [[n, [(i, i), (t_min, t_max)]]]

        for l in graph["pp_seg"].n[graph["pp_seg"]["from"] == to]:
            link = np.argwhere(newseg.index == str(l))[0][0]
            segments = segments + [[l, [(i, link), (t_max, t_max)]]]

    df = pd.DataFrame(segments)
    segments = df[0].unique()

    adata.uns["dendro_segments"] = dict(
        zip(segments, [df.loc[df[0] == s, 1].tolist() for s in segments])
    )

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "added \n"
        "    .obsm['X_dendro'], new embedding generated.\n"
        "    .uns['dendro_segments'] tree segments used for plotting."
    )


def hierarchy_pos(G, root=None, width=1.0, vert_gap=0.2, vert_loc=0, xcenter=0.5):

    """
    From Joel's answer at https://stackoverflow.com/a/29597209/2966723.
    Licensed under Creative Commons Attribution-Share Alike

    If the graph is a tree this will return the positions to plot this in a
    hierarchical layout.

    G: the graph (must be a tree)

    root: the root node of current branch
    - if the tree is directed and this is not given,
      the root will be found and used
    - if the tree is directed and this is given, then
      the positions will be just for the descendants of this node.
    - if the tree is undirected and not given,
      then a random choice will be used.

    width: horizontal space allocated for this branch - avoids overlap with other branches

    vert_gap: gap between levels of hierarchy

    vert_loc: vertical location of root

    xcenter: horizontal location of root
    """
    if not nx.is_tree(G):
        raise TypeError("cannot use hierarchy_pos on a graph that is not a tree")

    if root is None:
        if isinstance(G, nx.DiGraph):
            root = next(
                iter(nx.topological_sort(G))
            )  # allows back compatibility with nx version 1.11
        else:
            root = random.choice(list(G.nodes))

    def _hierarchy_pos(
        G, root, width=1.0, vert_gap=0.2, vert_loc=0, xcenter=0.5, pos=None, parent=None
    ):
        """
        see hierarchy_pos docstring for most arguments

        pos: a dict saying where all nodes go if they have been assigned
        parent: parent of this branch. - only affects it if non-directed

        """

        if pos is None:
            pos = {root: (xcenter, vert_loc)}
        else:
            pos[root] = (xcenter, vert_loc)
        children = list(G.neighbors(root))
        if not isinstance(G, nx.DiGraph) and parent is not None:
            children.remove(parent)
        if len(children) != 0:
            dx = width / len(children)
            nextx = xcenter - width / 2 - dx / 2
            for child in children:
                nextx += dx
                pos = _hierarchy_pos(
                    G,
                    child,
                    width=dx,
                    vert_gap=vert_gap,
                    vert_loc=vert_loc - vert_gap,
                    xcenter=nextx,
                    pos=pos,
                    parent=root,
                )
        return pos

    return _hierarchy_pos(G, root, width, vert_gap, vert_loc, xcenter)


### All functions bellow have been adapted from seaborn v0.11.1
### https://github.com/mwaskom/seaborn/blob/v0.11.1/seaborn/categorical.py


class _CategoricalPlotter(object):

    width = 0.8
    default_palette = "light"
    require_numeric = True

    def establish_variables(
        self,
        x=None,
        y=None,
        hue=None,
        data=None,
        orient=None,
        order=None,
        hue_order=None,
        units=None,
    ):
        """Convert input specification into a common representation."""

        x = data.get(x, x)
        y = data.get(y, y)
        hue = data.get(hue, hue)
        units = data.get(units, units)

        # Figure out the plotting orientation
        vals, groups = y, x
        # Put them into the common representation
        plot_data = [np.asarray(vals)]

        # Get a label for the value axis
        value_label = vals.name

        # Get the categorical axis label
        group_label = None
        if hasattr(groups, "name"):
            group_label = groups.name

        # Get the order on the categorical axis
        group_names = list(data.seg_pos.cat.categories)

        plot_data, value_label = self._group_longform(vals, groups, group_names)

        plot_hues = None
        hue_title = None
        plot_units = None

        self.orient = orient
        self.plot_data = plot_data
        self.group_label = group_label
        self.value_label = value_label
        self.group_names = group_names
        self.plot_hues = plot_hues
        self.hue_title = hue_title

    def _group_longform(self, vals, grouper, order):
        """Group a long-form variable by another with correct order."""

        # Group the val data
        grouped_vals = vals.groupby(grouper)
        out_data = []
        for g in order:
            try:
                g_vals = grouped_vals.get_group(g)
            except KeyError:
                g_vals = np.array([])
            out_data.append(g_vals)

        # Get the vals axis label
        label = vals.name

        return out_data, label


class _CategoricalScatterPlotter(_CategoricalPlotter):

    default_palette = "dark"
    require_numeric = False


class _SwarmPlotter(_CategoricalScatterPlotter):
    def __init__(
        self, x, y, hue, data, order, hue_order, dodge, orient, color, palette, n_jobs
    ):
        """Initialize the plotter."""
        self.establish_variables(x, y, hue, data, orient, order, hue_order)

        # Set object attributes
        self.dodge = dodge
        self.width = 0.8
        self.n_jobs = n_jobs

    def could_overlap(self, xy_i, swarm, d):
        """Return a list of all swarm points that could overlap with target.

        Assumes that swarm is a sorted list of all points below xy_i.
        """
        _, y_i = xy_i
        neighbors = []
        for xy_j in reversed(swarm):
            _, y_j = xy_j
            if (y_i - y_j) < d:
                neighbors.append(xy_j)
            else:
                break
        return np.array(list(reversed(neighbors)))

    def position_candidates(self, xy_i, neighbors, d):
        """Return a list of (x, y) coordinates that might be valid."""
        candidates = [xy_i]
        x_i, y_i = xy_i
        left_first = True
        for x_j, y_j in neighbors:
            dy = y_i - y_j
            dx = np.sqrt(max(d ** 2 - dy ** 2, 0)) * 1.05
            cl, cr = (x_j - dx, y_i), (x_j + dx, y_i)
            if left_first:
                new_candidates = [cl, cr]
            else:
                new_candidates = [cr, cl]
            candidates.extend(new_candidates)
            left_first = not left_first
        return np.array(candidates)

    def first_non_overlapping_candidate(self, candidates, neighbors, d):
        """Remove candidates from the list if they overlap with the swarm."""

        # IF we have no neighbours, all candidates are good.
        if len(neighbors) == 0:
            return candidates[0]

        neighbors_x = neighbors[:, 0]
        neighbors_y = neighbors[:, 1]

        d_square = d ** 2

        for xy_i in candidates:
            x_i, y_i = xy_i

            dx = neighbors_x - x_i
            dy = neighbors_y - y_i

            sq_distances = np.power(dx, 2.0) + np.power(dy, 2.0)

            # good candidate does not overlap any of neighbors
            # which means that squared distance between candidate
            # and any of the neighbours has to be at least
            # square of the diameter
            good_candidate = np.all(sq_distances >= d_square)

            if good_candidate:
                return xy_i

        # If `position_candidates` works well
        # this should never happen
        raise Exception(
            "No non-overlapping candidates found. " "This should not happen."
        )

    def beeswarm(self, orig_xy, d):
        """Adjust x position of points to avoid overlaps."""
        # In this method, ``x`` is always the categorical axis
        # Center of the swarm, in point coordinates
        midline = orig_xy[0, 0]

        # Start the swarm with the first point
        swarm = [orig_xy[0]]

        # Loop over the remaining points
        for xy_i in orig_xy[1:]:

            # Find the points in the swarm that could possibly
            # overlap with the point we are currently placing
            neighbors = self.could_overlap(xy_i, swarm, d)

            # Find positions that would be valid individually
            # with respect to each of the swarm neighbors
            candidates = self.position_candidates(xy_i, neighbors, d)

            # Sort candidates by their centrality
            offsets = np.abs(candidates[:, 0] - midline)
            candidates = candidates[np.argsort(offsets)]

            # Find the first candidate that does not overlap any neighbours
            new_xy_i = self.first_non_overlapping_candidate(candidates, neighbors, d)

            # Place it into the swarm
            swarm.append(new_xy_i)

        return np.array(swarm)

    def add_gutters(self, points, center, width):
        """Stop points from extending beyond their territory."""
        half_width = width / 2
        low_gutter = center - half_width
        off_low = points < low_gutter
        if off_low.any():
            points[off_low] = low_gutter
        high_gutter = center + half_width
        off_high = points > high_gutter
        if off_high.any():
            points[off_high] = high_gutter
        return points

    def swarm_points(self, ax, points, center, width, s, **kws):
        """Find new positions on the categorical axis for each point."""
        # Convert from point size (area) to diameter
        default_lw = mpl.rcParams["patch.linewidth"]
        lw = kws.get("linewidth", kws.get("lw", default_lw))
        dpi = ax.figure.dpi
        d = (np.sqrt(s) + lw) * (dpi / 72)
        # Transform the data coordinates to point coordinates.
        # We'll figure out the swarm positions in the latter
        # and then convert back to data coordinates and replot
        orig_xy = ax.transData.transform(points.get_offsets())

        # Order the variables so that x is the categorical axis

        # Do the beeswarm in point coordinates
        new_xy = self.beeswarm(orig_xy, d)

        # Transform the point coordinates back to data coordinates

        new_x, new_y = ax.transData.inverted().transform(new_xy).T

        # Add gutters

        self.add_gutters(new_x, center, width)

        # Reposition the points so they do not overlap
        return np.c_[new_x, new_y]

    def draw_swarmplot(self, ax, kws):
        """Plot the data."""
        s = kws.pop("s")

        centers = []
        swarms = []

        # Set the categorical axes limits here for the swarm math

        ax.set_xlim(-0.5, len(self.plot_data) - 0.5)

        # Plot each swarm
        cells = []
        for i, group_data in enumerate(self.plot_data):
            width = self.width
            swarm_data = np.asarray(group_data)

            # Sort the points for the beeswarm algorithm
            sorter = np.argsort(swarm_data)
            swarm_data = swarm_data[sorter]
            cell = group_data.index[sorter]

            # Plot the points in centered positions
            cat_pos = np.ones(swarm_data.size) * i
            points = ax.scatter(cat_pos, swarm_data, s=s, **kws)

            cells.append(cell)
            centers.append(i)
            swarms.append(points)

        # Autoscale the valus axis to set the data/axes transforms properly
        ax.autoscale_view(scalex=self.orient == "h", scaley=self.orient == "v")

        xys = ProgressParallel(
            n_jobs=self.n_jobs,
            total=len(self.plot_data),
            file=sys.stdout,
            desc="    segment ",
        )(
            delayed(self.swarm_points)(ax, swarm, center, 0.8, s, **kws)
            for center, swarm in zip(centers, swarms)
        )
        plt.close()
        dend = pd.concat(
            [pd.DataFrame(xy, index=cell) for cell, xy in zip(cells, xys)], axis=0
        )
        return dend

    def plot(self, ax, kws):
        """Make the full plot."""
        return self.draw_swarmplot(ax, kws)


def swarmplot(
    *,
    x=None,
    y=None,
    hue=None,
    data=None,
    order=None,
    hue_order=None,
    dodge=False,
    orient=None,
    color=None,
    palette=None,
    size=5,
    edgecolor="gray",
    linewidth=0,
    ax=None,
    n_jobs=1,
    **kwargs
):

    plotter = _SwarmPlotter(
        x, y, hue, data, order, hue_order, dodge, orient, color, palette, n_jobs
    )
    if ax is None:
        ax = plt.gca()

    kwargs.setdefault("zorder", 3)
    size = kwargs.get("s", size)
    if linewidth is None:
        linewidth = size / 10
    if edgecolor == "gray":
        edgecolor = "gray"
    kwargs.update(dict(s=size ** 2, edgecolor=edgecolor, linewidth=linewidth))

    dend = plotter.plot(ax, kwargs)
    return dend



from anndata import AnnData
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from joblib import delayed
from tqdm import tqdm
import sys
import igraph

from .utils import ProgressParallel

from .. import logging as logg
from .. import settings


def pseudotime(adata: AnnData, n_jobs: int = 1, n_map: int = 1, copy: bool = False):
    """\
    Compute pseudotime.

    Projects cells onto the tree, and uses distance from the root as a pseudotime value.

    Parameters
    ----------
    adata
        Annotated data matrix.
    n_jobs
        Number of cpu processes to use in case of performing multiple mapping.
    n_map
        number of probabilistic mapping of cells onto the tree to use. If n_map=1 then likelihood cell mapping is used.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    adata : anndata.AnnData
        if `copy=True` it returns or else add fields to `adata`:

        `.obs['edge']`
            assigned edge.
        `.obs['t']`
            assigned pseudotime value.
        `.obs['seg']`
            assigned segment of the tree.
        `.obs['milestone']`
            assigned region surrounding forks and tips.
        `.uns['pseudotime_list']`
            list of cell projection from all mappings.
    """

    if "root" not in adata.uns["graph"]:
        raise ValueError(
            "You need to run `tl.root` or `tl.roots` before projecting cells."
        )

    adata = adata.copy() if copy else adata

    graph = adata.uns["graph"]

    reassign, recolor = False, False
    if "milestones" in adata.obs:
        if adata.obs.milestones.dtype.name == "category":
            tmp_mil = adata.obs.milestones.cat.categories.copy()
            reassign = True
        if "milestones_colors" in adata.uns:
            tmp_mil_col = adata.uns["milestones_colors"].copy()
            recolor = True

    logg.info("projecting cells onto the principal graph", reset=True)

    if n_map == 1:
        df_l = [map_cells(graph, R=adata.obsm["X_R"], multi=False)]
    else:
        df_l = ProgressParallel(
            n_jobs=n_jobs, total=n_map, file=sys.stdout, desc="    mappings"
        )(
            delayed(map_cells)(graph=graph, R=adata.obsm["X_R"], multi=True)
            for m in range(n_map)
        )

    # formatting cell projection data
    for i in range(len(df_l)):
        df_l[i].index = adata.obs_names
    df_summary = df_l[0]

    df_summary["seg"] = df_summary["seg"].astype("category")
    df_summary["edge"] = df_summary["edge"].astype("category")

    # remove pre-existing palette to avoid errors with plotting
    if "seg_colors" in adata.uns:
        del adata.uns["seg_colors"]

    if set(df_summary.columns.tolist()).issubset(adata.obs.columns):
        adata.obs[df_summary.columns] = df_summary
    else:
        adata.obs = pd.concat([adata.obs, df_summary], axis=1)

    # list(map(lambda x: x.column))

    # todict=list(map(lambda x: dict(zip(["cells"]+["_"+s for s in x.columns.tolist()],
    #                                   [x.index.tolist()]+x.to_numpy().T.tolist())),df_l))
    names = np.arange(len(df_l)).astype(str).tolist()
    # vals = todict
    dictionary = dict(zip(names, df_l))
    adata.uns["pseudotime_list"] = dictionary

    if n_map > 1:
        adata.obs["t_sd"] = (
            pd.concat(
                list(
                    map(
                        lambda x: pd.Series(x["t"]),
                        list(adata.uns["pseudotime_list"].values()),
                    )
                ),
                axis=1,
            )
            .apply(np.std, axis=1)
            .values
        )

    milestones = pd.Series(index=adata.obs_names)
    for seg in graph["pp_seg"].n:
        cell_seg = adata.obs.loc[adata.obs["seg"] == seg, "t"]
        if len(cell_seg) > 0:
            milestones[
                cell_seg.index[
                    (cell_seg - min(cell_seg) - (max(cell_seg - min(cell_seg)) / 2) < 0)
                ]
            ] = graph["pp_seg"].loc[int(seg), "from"]
            milestones[
                cell_seg.index[
                    (cell_seg - min(cell_seg) - (max(cell_seg - min(cell_seg)) / 2) > 0)
                ]
            ] = graph["pp_seg"].loc[int(seg), "to"]
    adata.obs["milestones"] = milestones
    adata.obs.milestones = (
        adata.obs.milestones.astype(int).astype("str").astype("category")
    )

    adata.uns["graph"]["milestones"] = dict(
        zip(
            adata.obs.milestones.cat.categories,
            adata.obs.milestones.cat.categories.astype(int),
        )
    )

    while reassign:
        if "tmp_mil_col" not in locals():
            break
        if len(tmp_mil_col) != len(adata.obs.milestones.cat.categories):
            break
        rename_milestones(adata, tmp_mil)
        if recolor:
            adata.uns["milestones_colors"] = tmp_mil_col
        reassign = False

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "added\n"
        "    .obs['edge'] assigned edge.\n"
        "    .obs['t'] pseudotime value.\n"
        "    .obs['seg'] segment of the tree assigned.\n"
        "    .obs['milestones'] milestone assigned.\n"
        "    .uns['pseudotime_list'] list of cell projection from all mappings."
    )

    return adata if copy else None


def map_cells(graph, R, multi=False):
    import igraph

    g = igraph.Graph.Adjacency((graph["B"] > 0).tolist(), mode="undirected")
    # Add edge weights and node labels.
    g.es["weight"] = graph["B"][graph["B"].nonzero()]
    if multi:
        rrm = (
            np.apply_along_axis(
                lambda x: np.random.choice(np.arange(len(x)), size=1, p=x),
                axis=1,
                arr=R,
            )
        ).T.flatten()
    else:
        rrm = np.apply_along_axis(np.argmax, axis=1, arr=R)

    def map_on_edges(v):
        vcells = np.argwhere(rrm == v)

        if vcells.shape[0] > 0:
            nv = np.array(g.neighborhood(v, order=1))
            nvd = np.array(g.shortest_paths(v, nv)[0])

            spi = np.apply_along_axis(np.argmax, axis=1, arr=R[vcells, nv[1:]])
            ndf = pd.DataFrame(
                {
                    "cell": vcells.flatten(),
                    "v0": v,
                    "v1": nv[1:][spi],
                    "d": nvd[1:][spi],
                }
            )

            p0 = R[vcells, v].flatten()
            p1 = np.array(
                list(map(lambda x: R[vcells[x], ndf.v1[x]], range(len(vcells))))
            ).flatten()

            alpha = np.random.uniform(size=len(vcells))
            f = np.abs(
                (np.sqrt(alpha * p1 ** 2 + (1 - alpha) * p0 ** 2) - p0) / (p1 - p0)
            )
            ndf["t"] = (
                graph["pp_info"].loc[ndf.v0, "time"].values
                + (
                    graph["pp_info"].loc[ndf.v1, "time"].values
                    - graph["pp_info"].loc[ndf.v0, "time"].values
                )
                * alpha
            )
            ndf["seg"] = 0
            isinfork = (graph["pp_info"].loc[ndf.v0, "PP"].isin(graph["forks"])).values
            ndf.loc[isinfork, "seg"] = (
                graph["pp_info"].loc[ndf.loc[isinfork, "v1"], "seg"].values
            )
            ndf.loc[~isinfork, "seg"] = (
                graph["pp_info"].loc[ndf.loc[~isinfork, "v0"], "seg"].values
            )

            return ndf
        else:
            return None

    df = list(map(map_on_edges, range(graph["B"].shape[1])))
    df = pd.concat(df)
    df.sort_values("cell", inplace=True)
    # df.index = graph["cells_fitted"]

    df["edge"] = df.apply(lambda x: str(int(x[1])) + "|" + str(int(x[2])), axis=1)

    df.drop(["cell", "v0", "v1", "d"], axis=1, inplace=True)

    return df


def rename_milestones(adata, new, copy: bool = False):

    adata = adata.copy() if copy else adata

    adata.uns["graph"]["milestones"] = dict(
        zip(new, list(adata.uns["graph"]["milestones"].values()))
    )

    adata.obs.milestones = adata.obs.milestones.cat.rename_categories(new)

    return adata if copy else None
